#ifndef HATPPreprocessor_HH
#define HATPPreprocessor_HH

#include <iostream>
#include <fstream>
#include <algorithm>
#include <string>
#include <utility>

#include <cstdio>

#include <boost/regex.hpp>

class Range{
private:
	size_t beginComplete, endComplete;
	size_t beginReal, endReal;
	std::string realFilename;
	
	Range();
	
public:
	Range(size_t beginComplete, size_t beginReal, const std::string& realFilename): beginComplete(beginComplete), endComplete(beginComplete), beginReal(beginReal), endReal(beginReal), realFilename(realFilename) {}
	
	Range(const Range& range): beginComplete(range.beginComplete), endComplete(range.endComplete), beginReal(range.beginReal), endReal(range.endReal), realFilename(range.realFilename) {}
	
	Range& operator=(Range range){
		std::swap(*this, range);
		return *this;
	}
	
	void increaseLineNumber(){
		++endComplete;
		++endReal;
	}
	
	size_t lastLine() const {
		return endComplete;
	}
	
	std::pair<std::string, size_t> getLine(size_t line) const {
		if(line<beginComplete || line>endComplete){
			return std::pair<std::string, size_t>("", 0);
		}
		
		size_t offset=line-beginComplete;
		
		if(beginReal+offset>endReal){
			return std::pair<std::string, size_t>("", 0);
		}
		return std::make_pair(realFilename, beginReal+offset);
	}
	
	bool operator==(size_t line) const {
		return(line>=beginComplete&&line<=endComplete);
	}
};

class Preprocessor{
private:
	std::string path;
	std::ofstream complete;
	std::ifstream completeFile;
	std::string completeDomainFilename;
	std::vector<Range> content; //Associate to each line from the complete domain a line fron the real file the line comes from.
	bool worked;
	
	Preprocessor();
	Preprocessor(const Preprocessor&);
	Preprocessor& operator=(const Preprocessor&);

public:
	Preprocessor(const std::string& domainFile): completeDomainFilename(domainFile+".complete") {
		complete.open(completeDomainFilename.c_str(), std::ios_base::in | std::ios_base::out | std::ios_base::trunc);
		
		if(!complete.is_open()){
			std::cerr<<"Error while preparing the temporary file : "<<completeDomainFilename<<std::endl;
			worked=false;
		}else{
			size_t length=domainFile.rfind("/");
			if(length!=0){
				path=domainFile.substr(0, length+1);
			}
		
			worked=includeFile(domainFile);
		
			complete.close();
		
			if(worked){
				completeFile.open(completeDomainFilename.c_str());
				if(!completeFile.is_open()){
					worked=false;
				}
			}
		}
	}
	
	~Preprocessor(){
		remove(completeDomainFilename.c_str());
		completeFile.close();
	}
	
	bool isReady(){
		return worked;
	}

private:
	bool includeFile(const std::string& includeName){
		std::ifstream include;
		include.open(includeName.c_str());
	
		if(!include.is_open()){
			std::cerr<<"Problem while openning the file \""<<includeName<<"\""<<std::endl;
			return false;
		}
	
		size_t lineNumber=1;
		bool fileLinesStored=false;
		bool newIncludedFile=false;
		std::string line;
		size_t it;
		while(!include.eof()){
			std::getline(include, line);
			if(include.eof()){ continue; }
			it=line.find("#include");
		
			if(it!=std::string::npos){
				boost::cmatch what;
				if(boost::regex_match(line.c_str(), what, boost::regex("[[:space:]]*[[.number-sign.]]include[[:space:]]\".*\"[[:space:]]*"))){
					size_t first=line.find("\"")+1;
					size_t last=line.rfind("\"");
					size_t length=last-first;
					if(length!=0){
						std::string includeFilename=line.substr(first, length);
						if(!includeFile(path+includeFilename)){
							std::cerr<<includeName<<":"<<lineNumber<<" Error : Impossible to include \""<<includeFilename<<"\""<<std::endl;
							return false;
						}
						newIncludedFile=true;
					}else{
						std::cerr<<includeName<<":"<<lineNumber<<" Error : Wrong included file"<<std::endl;
						return false;
					}
				}else{
					std::cerr<<includeName<<":"<<lineNumber<<" Error : Include has wrong format"<<std::endl;
					return false;
				}
			}else{
				complete<<line<<std::endl;
				if(!fileLinesStored){
					if(content.empty()){
						content.push_back(Range(1, lineNumber, includeName));
					}else{
						size_t previous=content.back().lastLine()+1;
						content.push_back(Range(previous, lineNumber, includeName));
					}
					fileLinesStored=true;
					newIncludedFile=false;
				}else{
					if(newIncludedFile){
						size_t previous=content.back().lastLine()+1;
						content.push_back(Range(previous, lineNumber, includeName));
						fileLinesStored=true;
						newIncludedFile=false;
					}else{
						content.back().increaseLineNumber();
					}
				}
			}
			++lineNumber;
		}
		include.close();
	
		return true;
	}

public:	
	std::ifstream& getCompleteDomain(){
		return completeFile;
	}
	
	std::pair<std::string, size_t> getLine(size_t line) const {
		std::vector<Range>::const_iterator it;
		it=std::find(content.begin(), content.end(), line);
		if(it!=content.end()){
			return it->getLine(line);
		}
		return std::pair<std::string, size_t>("", 0);
	}
};

#endif //HATPPreprocessor_HH
