/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>
#include<sstream>

#include "PlanTree.hh"
#include "Substitution.hh"
#include "LightWorldBase.hh"

using namespace std;

/**
 * constructor by copying, we make a copy of all nodes
 */
PlanTree::PlanTree(PlanTree const& PT)
{
	unsigned long int i,j;
	vector<PlanTreeTaskList*> const& PTnodes = PT.getNodes();
	map<PlanTreeTaskList*,PlanTreeTaskList*> oldAndNewAdresses;
	PlanTreeTaskList* PTTL;
	for(i=0 ; i<PTnodes.size() ; ++i) {
		PTTL = new PlanTreeTaskList(*PTnodes[i]);
		PlanTreeNodes.push_back(PTTL);
		// we keep correspondance between PT adresses and current adresses
		oldAndNewAdresses[PTnodes[i]] = PTTL;
	}
	// we update rootList
	rootList = oldAndNewAdresses[PT.getRootList()];
	// we update decompositions of each nodes member
	vector<PlanTreeTaskList*> PTdecompos, decompos;
	for(i=0 ; i<PTnodes.size() ; ++i) {
		PTdecompos = PTnodes[i]->getDecompositions();
		decompos.clear();
		for(j=0 ; j<PTdecompos.size() ; ++j) {
			if(PTdecompos[j] == NULL)
				decompos.push_back(NULL);
			else
				decompos.push_back(oldAndNewAdresses[PTdecompos[j]]);
		}
		PlanTreeNodes[i]->setDecompositions(decompos);
	}
}

/**
 * return a vector containing multiple copies (vector size = nbCopies) of the
 * current PlanTree. This function is useful to save time during planning process
 * where we make a lot of copies of the PlanTree. Call rootList->getMultipleCopies(nbCopies).
 */
vector<PlanTree*> PlanTree::getMultipleCopies(unsigned long int nbCopies) const
{
	unsigned long int i,j,k;
	PlanTreeTaskList* PTTL;
	vector<vector<PlanTreeTaskList*> > copiesNodes(nbCopies, vector<PlanTreeTaskList*>(PlanTreeNodes.size()));
	vector<map<PlanTreeTaskList*, PlanTreeTaskList*> > copiesAdressesCorrespondance(nbCopies, map<PlanTreeTaskList*, PlanTreeTaskList*>());
	// we complete copiesNodes and copiesAdressesCorrespondance
	for(j=0 ; j<nbCopies ; ++j) {
		for(i=0 ; i<PlanTreeNodes.size() ; ++i) {
			PTTL = new PlanTreeTaskList(*PlanTreeNodes[i]);
			copiesNodes[j][i]=PTTL;
			copiesAdressesCorrespondance[j][PlanTreeNodes[i]] = PTTL;
		}
		//copiesAdressesCorrespondance[j][NULL] = NULL;
	}
	// we update rootLists and decompositions of each copies
	vector<PlanTreeTaskList*> decompos;
	for(i=0 ; i<PlanTreeNodes.size() ; ++i) {
		vector<PlanTreeTaskList*> const& currentDecompos = PlanTreeNodes[i]->getDecompositions();
		decompos.resize(currentDecompos.size());
		for(j=0 ; j<nbCopies ; ++j) {
			for(k=0 ; k<currentDecompos.size() ; ++k)
				decompos[k] = copiesAdressesCorrespondance[j][currentDecompos[k]];
			copiesNodes[j][i]->setDecompositions(decompos);
		}
	}
	// we create PlanTrees
	vector<PlanTree*> vecExit(nbCopies);
	for(i=0 ; i<nbCopies ; ++i)
		vecExit[i] = new PlanTree(copiesAdressesCorrespondance[i][rootList], copiesNodes[i]);
	return vecExit;
}

/**
 * returns a copy of the current PlanTree and the corresponding resulting TaskList
 * (this is done by a hierarchical exploration of the Tree)
 */
pair<TaskList*, PlanTree*> PlanTree::makeCopy()
{
	map<unsigned long int, TaskList*> TLstock;
	map<unsigned long int, unsigned long int> oldAndNewIDs;
	// root list will be associated to EMPTY_ID
	rootList->getNodesInTaskListForms(EMPTY_ID, TLstock, oldAndNewIDs);
	// we start from root list and we decompose its nodes to make a copy
	map<unsigned long int, bool> TLused; // to know TaskList already used
	map<unsigned long int, bool>::iterator ITused;
	map<unsigned long int, TaskList*>::iterator itTL;
	for(itTL = TLstock.begin() ; itTL != TLstock.end() ; ++itTL)
		TLused[itTL->first] = false;
	TaskList* TL = TLstock[EMPTY_ID];
	TLused[EMPTY_ID] = true;
	PlanTree* PT = new PlanTree(TaskList(*TL, false));
	bool allDone = false;
	do {
		// we explore TLstock to update TL and PT
		for(itTL = TLstock.begin() ; itTL != TLstock.end() ; ++itTL)
			// the tasklist has not been already used and can be used
			if( !(TLused[itTL->first]) && PT->containsNode(oldAndNewIDs[itTL->first]) ) {
				PT->decompose(oldAndNewIDs[itTL->first], *(itTL->second));
				TL->substitute(oldAndNewIDs[itTL->first], itTL->second);
				// we free memory
				itTL->second->noNeedToDeleteNodes();
				delete itTL->second;
				TLused[itTL->first] = true;// we update TLused
			}
		// we update allDone
		allDone = true;
		for(ITused = TLused.begin() ; ITused != TLused.end() ; ++ITused)
			if(ITused->second == false) {
				allDone = false;
				break;
			}
	} while(!allDone);
	return make_pair(TL, PT);
}

/**
 * return true if node with ID = argID is found somewhere in the Tree
 */
bool PlanTree::containsNode(unsigned long int argID)
{
	if(rootList == NULL)
		cerr << "ERROR: when trying to check if node with ID " << argID << " is in the Tree, the root node is empty" << endl;
	else { // we explore nodes checking IDs of PlanTreeTaskLists
		for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i)
			if(PlanTreeNodes[i]->containsNode(argID))
				return true;
	}
	return false;
}


/**
 * return a pointer on the decomposition associated to node with the ID id,
 * if id is not found return NULL
 */
PlanTreeTaskList* PlanTree::getDecompositionOfID(unsigned long int argID) const
{
	for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i)
		if(PlanTreeNodes[i]->containsNode(argID))
			return PlanTreeNodes[i]->getDecompositionOfID(argID);
	return NULL;
}

/**
 * try to find argID in the tree, if it is found create a new PlanTreeTaskList from TL and add it in
 * the Tree
 */
void PlanTree::decompose(unsigned long int argID, TaskList const& TL)
{
	if(rootList == NULL) // if there is no root node
		cerr << "ERROR: when trying to make a decomposition in PlanTree, the root node is empty" << endl;
	else {
	// we explore nodes to find argID, if we found it we make a PlanTreeTaskList from TL and we add it in nodes
		PlanTreeTaskList* PTfather = NULL;
		for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i) // we explore nodes to find argID
			if(PlanTreeNodes[i]->containsNode(argID)) {
				PTfather = PlanTreeNodes[i];
				break;
			}

		if(PTfather == NULL) {// we have not found argID in nodes
			ostringstream st;
			st << argID;
			cerr << "ERROR: when trying to make a decomposition in PlanTree, no node with ID=";
			cerr << st.str() << endl;
		}
		else {// everyting OK
			PlanTreeTaskList* PT = new PlanTreeTaskList(TL, PTfather->getLevel()+1);
			PlanTreeNodes.push_back(PT);
			PTfather->decompose(argID, PT); // we update decompositions of PTfather
		}
	}
}

/**
 * replace a given node by another
 */
bool PlanTree::replaceNode(unsigned long int argID, TaskListNode const& Tln){
	if(rootList == NULL){
		cerr << "ERROR: when trying to make a decomposition in PlanTree, the root node is empty" << endl;
		return false;
	}

	//We explore all nodes to find argID, if found a PlanTreeTaskList is created from TL and added to the list of nodes
	PlanTreeTaskList* PTfather = NULL;
	
	//Find the father of argID
	for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i){
		if(PlanTreeNodes[i]->containsNode(argID)) {
			PTfather = PlanTreeNodes[i];
			break;
		}
	}
	
	//If the node does not exists
	if(PTfather == NULL) {
		cerr << "ERROR: when trying to make a decomposition in PlanTree, no node with ID=" << argID << endl;
		return false;
	}
	
	//The node exists, so it is replaced
	return PTfather->replaceNode(argID, Tln);
}

/**
 * calls rootList->toString(ROOT_LEVEL) and returns the result
 * (this is done by a hierarchical exploration of the Tree)
 */
string PlanTree::toString() const
{
	string s;
	if(rootList != NULL) s = rootList->toString();
	else cerr << "ERROR: when trying to convert PlanTree in a string, the root node is empty" << endl;
	return s;
}

/**
 * call rootList->eraseEmptyBranches() to erase all branches of the tree resulting in empty TaskLists
 * (this is done by a hierarchical exploration of the Tree)
 */
void PlanTree::eraseEmptyBranches()
{
	rootList->eraseEmptyBranches();
	// we explore PlanTreeNodes and we erase all empty PlanTreeTaskList to free memory
	for(vector<PlanTreeTaskList*>::iterator it = PlanTreeNodes.begin() ; it != PlanTreeNodes.end() ;) {
		if( (*it)->isEmpty() ) {
			delete *it;
			it = PlanTreeNodes.erase(it);
		}
		else it++;
	}
}

/**
 * call rootList->propagateInvolvedAgents(TL) to propagate who is involved in all tasks in the tree according TL
 * (TL must be AllDone)
 */
void PlanTree::propagateInvolvedAgent()
{
	// we explore PlanTreeNodes to affect involved agents for all leaves
	vector<PlanTreeTaskListNode*> vecNodes;
	vector<bool> vecLeaves;
	unsigned long int i,j;
	for(i=0 ; i<PlanTreeNodes.size() ; ++i) {
		vecNodes = PlanTreeNodes[i]->getNodes();
		vecLeaves = PlanTreeNodes[i]->getAreLeaves();
		for(j=0 ; j<vecNodes.size() ; ++j)
			if(vecLeaves[j])
				vecNodes[j]->updateInvolvedAgents();
	}
	rootList->propagateInvolvedAgents();
}

/**
 * converts the current PlanTree to the abstraction of the plan
 * explore hierarchicaly the Tree and cut all branches with only one associated agent
 */
void PlanTree::convertToAbstraction()
{
	// we explore PlanTreeNodes and we erase all PlanTreeTaskList with only one Agent involved because they will be deleted
	// by function rootList->cutPrivateBranches()
	for(vector<PlanTreeTaskList*>::iterator it = PlanTreeNodes.begin() ; it != PlanTreeNodes.end() ;) {
		if( (*it)->getNbInvolvedAgents()==1 && *it != rootList ) // we do not erase the root list
			it = PlanTreeNodes.erase(it);
		else it++;
	}
	rootList->cutPrivateBranches(); // explore the Tree and cuts all branches which corresponds to private level
}


/**
 * return all PlanTreeTaskListNode of the current PlanTree using task called taskName
 */
vector<PlanTreeTaskListNode*> PlanTree::find(string const& taskName) const
{
	unsigned long int i,j;
	vector<PlanTreeTaskListNode*> vecExit;
	for(i=0 ; i<PlanTreeNodes.size() ; ++i) {
		vector<PlanTreeTaskListNode*> const& elementaryNodes = PlanTreeNodes[i]->getNodes();
		for(j=0 ; j<elementaryNodes.size() ; ++j)
			if(elementaryNodes[j]->getInstance().getGeneralTask().getName() == taskName)
				vecExit.push_back(elementaryNodes[j]);
	}
	return vecExit;
}

/**
 * explore PlanTreeNodes and return the highest level found
 */
unsigned long int PlanTree::getDepth() const
{
	unsigned long int depth=0;
	for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i)
		if(PlanTreeNodes[i]->getLevel() > depth)
			depth = PlanTreeNodes[i]->getLevel();
	// NB: a tree will only a root node has a depth at 0
	return depth;
}

/**
 * return true if starting at node with ID IDstart we can reach 'end' using
 * decomposition links
 */
bool PlanTree::isThereAPathFromTo(unsigned long int IDstart, PlanTreeTaskList* end) const
{
	bool modif;
	unsigned long int i,j;
	vector<PlanTreeTaskList*> decompo;
	map<PlanTreeTaskList*, unsigned long int> PTTLordersInVector;
	PlanTreeTaskList* start = this->getDecompositionOfID(IDstart);
	if(start == NULL)
		return false;
	if( end == NULL ) {
		cerr << "ERROR: trying to find a path between two nodes in a PlanTree,";
		cerr << " the destination node is NULL" << endl;
	}
	else {
		// we create a map associating PlanTreeTaskList* and their order in the vector
		for(i=0 ; i<PlanTreeNodes.size() ; ++i)
			PTTLordersInVector[PlanTreeNodes[i]] = i;
		// we init connected
		vector<bool> connected = vector<bool>(PlanTreeNodes.size(), false);
		connected[PTTLordersInVector[start]] = true;
		do {
			modif = false;
			for(i=0 ; i<PlanTreeNodes.size() ; ++i)
				if(connected[PTTLordersInVector[PlanTreeNodes[i]]]) {
				// if the nodes is connected all its decompsoitions are connected too
					decompo = PlanTreeNodes[i]->getDecompositions();
					for(j=0 ; j<decompo.size() ; ++j)
						if( decompo[j]!=NULL && !connected[PTTLordersInVector[decompo[j]]] ) {
							connected[PTTLordersInVector[decompo[j]]] = true;
							modif = true;
						}
				}
			if(connected[PTTLordersInVector[end]])
				return true;
		} while(modif);
	}
	// propagation in connected achieved and TLend is not connected
	return false;
}

/**
 * return true if the PlanTree described by subPTnodes and
 * hierarchical_links is a part of the current PlanTree
 * NB: if members of hierarchical_links are not the same as those in
 *		 subPTnodes, this will lead to a system crash
 */
unsigned long int PlanTree::checkSubTree(vector<TaskList*> const& subPTnodes,
	map<TaskListNode*, pair<TaskList*, string> > const& hierarchical_links) const
{
	unsigned long int i,j,k,l;
	// used to get correspondence of TaskListNodes and TaskList in subPTnodes
	map<TaskListNode*, TaskList*> TLnodesToTL;
	vector<TaskListNode*> vecTLnodes;
	for(i=0 ; i<subPTnodes.size() ; ++i)
	{
		vecTLnodes = subPTnodes[i]->getNodes();
		for(j=0 ; j<vecTLnodes.size() ; ++j)
			TLnodesToTL[vecTLnodes[j]] = subPTnodes[i];
	}

	// we compute all possible combinations using PlanTreeTaskList::getSubListCorrespondence
	vector< map<TaskListNode*, PlanTreeTaskListNode*> > combinations(0), tmp_combinations, candidates;
	vector< map<TaskList*, PlanTreeTaskList*> > listCorrespondence(0), tmp_listCorrespondence;
	map<TaskListNode*, PlanTreeTaskListNode*> current_combination;
	map<TaskList*, PlanTreeTaskList*> current_listCorrespondence;
	bool correspendenceFound;
	for(i=0 ; i<subPTnodes.size() ; ++i)
	{
		correspendenceFound = false;
		tmp_combinations.clear();
		tmp_listCorrespondence.clear();

		for(j=0 ; j<PlanTreeNodes.size() ; ++j)
		{
			candidates = PlanTreeNodes[j]->getSubListCorrespondence(*subPTnodes[i]);

			// if there is at least one potential candidate, we complete combinations and listCorrespondence
			if( !candidates.empty() )
			{
				correspendenceFound = true;

				if(i == 0) // initialization
					for(k=0 ; k<candidates.size() ; ++k){
						tmp_combinations.push_back(candidates[k]);

						current_listCorrespondence = map<TaskList*, PlanTreeTaskList*>();
						current_listCorrespondence[subPTnodes[i]] = PlanTreeNodes[j];
						tmp_listCorrespondence.push_back(current_listCorrespondence);
					}
				else // we compute tmp_combinations from combinations and candidates
					for(k=0 ; k<candidates.size() ; ++k)
						for(l=0 ; l<combinations.size() ; ++l){
							current_combination = combinations[l];
							current_combination.insert(candidates[k].begin(), candidates[k].end());
							tmp_combinations.push_back(current_combination);

							current_listCorrespondence = listCorrespondence[l];
							current_listCorrespondence[subPTnodes[i]] = PlanTreeNodes[j];
							tmp_listCorrespondence.push_back(current_listCorrespondence);
						}
			}
		}

		combinations = tmp_combinations;
		listCorrespondence = tmp_listCorrespondence;

		if( !correspendenceFound ) // no correspondence found in the PlanTree for subPTnodes[i]
			return 0;
	}

	// for all members in combinations we check the hierarchical links
	bool good_combination;
	unsigned long int nb=0;
	map<TaskListNode*, pair<TaskList*, string> >::const_iterator it;
	for(i=0 ; i<combinations.size() ; ++i)
	{
		good_combination = true;

		for(it = hierarchical_links.begin() ; it != hierarchical_links.end() && good_combination ; ++it)
		{
			if( (it->second).second == IMMEDIATE_DECOMPOSITION ) {
				// we check if (it->second).first is a decomposition of it->first
				if( listCorrespondence[i][TLnodesToTL[it->first]]->getDecompositionOfID(combinations[i][it->first]
																																								->getID()) != listCorrespondence[i][(it->second).first] )
				{
					good_combination = false;
					break;
				}
			}
			else if( (it->second).second == DEEP_DECOMPOSITION ) {
				// we check if there is a path in the PlanTree from it->first to (it->second).first
				if( !(this->isThereAPathFromTo(combinations[i][it->first]->getID(),
																			 listCorrespondence[i][(it->second).first])) )
				{
					good_combination = false;
					break;
				}
			}
			else // error link type is not recognized
			{
				cerr << "ERROR when checking if a PLanTree contains a subPlanTree description, link type \"";
				cerr << (it->second).second << "\" unrecognized" << endl;
				return 0;
			}
		}
		
		if(good_combination)
			nb++;
	}

	// no combination OK so we return false
	return nb;
}

/**
 * explore PlanTreeNodes and return all possible Subsitutions according subPTnodes
 * arguments
 */
vector<Substitution*> PlanTree::getSubstitutionsOnNodes(vector<TaskList*> const& subPTnodes,
	vector<string> const& varNames) const
{
	unsigned long int i,j,k,l,n;
	vector<PlanTreeTaskListNode*> correspondingNodes;
	vector<Substitution*> vecSub, tmpVecSub;
	Substitution* subst;
	bool compatible;

	for(i=0 ; i<subPTnodes.size() ; ++i)
	{
		vector<TaskListNode*> const& PTelementaryNodes = subPTnodes[i]->getNodes();
		for(j=0 ; j<PTelementaryNodes.size() ; ++j)
		{
			correspondingNodes = this->find(PTelementaryNodes[j]->getInstance().getGeneralTask().getName());

			if(correspondingNodes.empty()) // it misses a node we return an empty vector
			{
				for(k=0 ; k<vecSub.size() ; ++k)
					delete vecSub[k];

				return vector<Substitution*>(0);
			}
			else // we have found at least one corresponding node in current PlanTree
			{
				vector<string> const& vecElementaryNodeArgs = PTelementaryNodes[j]->getInstance().getArgs();

				if(vecSub.empty()) //initialization of vecSub
				{
					for(l=0 ; l<correspondingNodes.size() ; ++l)
					{
						subst = new Substitution(varNames);
						vector<string> const& vecArgs = correspondingNodes[l]->getInstance().getArgs();
						compatible = true;

						// we are sure that vecElementaryNodeArgs and vecArgs have the same size
						// because for a given task there is a fixed number of arguments
						for(k=0 ; k<vecArgs.size() ; ++k)
						{
							if( !(subst->isInstancied(vecElementaryNodeArgs[k])) )
					subst->setValue(vecElementaryNodeArgs[k], vecArgs[k]);
							else if( subst->getValue(vecElementaryNodeArgs[k]) != vecArgs[k] )
							{
								delete subst;
								compatible = false;
								break;
							}
						}

						if(compatible)
							tmpVecSub.push_back(subst);
					}
				} // end initialization of vecSub
				else // vecSub is not empty
				{
					for(n=0 ; n<vecSub.size() ; ++n)
					{
						for(l=0 ; l<correspondingNodes.size() ; ++l)
						{
							subst = new Substitution(*vecSub[n]);
							vector<string> const& vecArgs2 = correspondingNodes[l]->getInstance().getArgs();
							compatible = true;

							// we are sure that vecElementaryNodeArgs and vecArgs have the same size
							// because for a given task there is a fixed number of arguments
							for(k=0 ; k<vecArgs2.size() ; ++k)
							{
								if( !(subst->isInstancied(vecElementaryNodeArgs[k])) )
									subst->setValue(vecElementaryNodeArgs[k], vecArgs2[k]);
								else if( subst->getValue(vecElementaryNodeArgs[k]) != vecArgs2[k] )
								{
									delete subst;
									compatible = false;
									break;
								}
							}

							if(compatible)
					tmpVecSub.push_back(subst);
						} // end for loop to explore correspondingNodes
					} // end for loop to explore vecSub
				} // end vecSub is not empty
			} // end there is one corresponding node in current PlanTree
		} // end for loop to explore PTelementaryNodes

		for(k=0 ; k<vecSub.size() ; ++k)
			delete vecSub[k];

		if(tmpVecSub.empty()) // no Substitution available we return an empty vector
			return vector<Substitution*>(0);
		else
		{
			vecSub = tmpVecSub;
			tmpVecSub.clear();
		}
	} // end for loop to explore subPTnodes

	return vecSub;
}

/**
 * check if the tree described by subPTnodes and hierarchical_links
 * is detected in the current PlanTree
 * i.e. all nodes of subPTnodes can be found in the current Plantree and all link
 * types are respected (hierarchical_links must contains only links of type
 * IMMEDIATE_DECOMPOSITION and DEEP_DECOMPOSITION)
 * Variables in subPTnodes must be in varNames
 * return a vector containing valid Substitution associated to the number of patterns
 * recognized in the current Plantree using it.
 */
vector<SocialRuleData*> PlanTree::checkSubTree2(LightWorldBase const& wb,
	vector<Substitution*> const& vecSub, vector<TaskList*> const& subPTnodes,
	map<TaskListNode*, pair<TaskList*,string> > const& hierarchical_links,
	vector<string> varNames) const
{
	//for all possible Substitution in vecSub, we check links
	vector<SocialRuleData*> vecExit;
	unsigned long int i,j,nb;
	TaskList* Rebuiltnode;
	vector<TaskList*> RebuiltSubTLnodes;
	map<TaskListNode*, pair<TaskList*, string> > RebuiltHierarchicalLinks;
	map<TaskListNode*, pair<TaskList*, string> >::const_iterator it;
	map<TaskListNode*, TaskListNode*> oldToRebuiltElementaryNodes;
	map<TaskList*, TaskList*> oldToRebuiltNodes;

	for(i=0 ; i<vecSub.size() ; ++i)
	{
		RebuiltSubTLnodes.clear();
		RebuiltHierarchicalLinks.clear();
		oldToRebuiltElementaryNodes.clear();
		oldToRebuiltNodes.clear();

		// we rebuild subPTnodes according vecSub[i]
		for(j=0 ; j<subPTnodes.size() ; ++j)
		{
			Rebuiltnode = subPTnodes[j]->Rebuild(*vecSub[i], wb, oldToRebuiltElementaryNodes);
			RebuiltSubTLnodes.push_back(Rebuiltnode);
			oldToRebuiltNodes[subPTnodes[j]] = Rebuiltnode;
		}

		// we update hierarchical_links
		for(it = hierarchical_links.begin() ; it != hierarchical_links.end() ; ++it)
			RebuiltHierarchicalLinks[oldToRebuiltElementaryNodes[it->first]] = make_pair(oldToRebuiltNodes[(it->second).first], (it->second).second);

		nb = this->checkSubTree(RebuiltSubTLnodes, RebuiltHierarchicalLinks);
		if(nb > 0)
			vecExit.push_back(new SocialRuleData(vecSub[i],nb));
		else
			delete vecSub[i];

		// we free memory
		for(j=0 ; j<RebuiltSubTLnodes.size() ; ++j)
			delete RebuiltSubTLnodes[j];
	} // end for loop to explore vecSub

	return vecExit;
}

/**
 * assignment operator (make a copy of PT)
 */
PlanTree& PlanTree::operator=(PlanTree const& PT)
{
	// we erase members of PlanTreeNodes
	unsigned long int i;
	for(i=0 ; i<PlanTreeNodes.size() ; ++i)
		delete PlanTreeNodes[i];
	PlanTreeNodes.clear();

	// we complete PlanTreeNodes with copies of PT nodes
	unsigned long int j;
	vector<PlanTreeTaskList*> PTnodes = PT.getNodes();
	map<PlanTreeTaskList*,PlanTreeTaskList*> oldAndNewAdresses;
	PlanTreeTaskList* PTTL;

	for(i=0 ; i<PTnodes.size() ; ++i)
	{
		PTTL = new PlanTreeTaskList(*PTnodes[i]);
		PlanTreeNodes.push_back(PTTL);
		// we keep correspondance between PT adresses and current adresses
		oldAndNewAdresses[PTnodes[i]] = PTTL;
	}

	// we update rootList
	rootList = oldAndNewAdresses[PT.getRootList()];

	// we update decompositions of each nodes member
	vector<PlanTreeTaskList*> PTdecompos, decompos;
	for(i=0 ; i<PTnodes.size() ; ++i)
	{
		PTdecompos = PTnodes[i]->getDecompositions();
		decompos.clear();

		for(j=0 ; j<PTdecompos.size() ; ++j)
		{
			if(PTdecompos[j] == NULL)
				decompos.push_back(NULL);
			else
				decompos.push_back(oldAndNewAdresses[PTdecompos[j]]);
		}

		PlanTreeNodes[i]->setDecompositions(decompos);
	}

	return *this;
}
