/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_PLAN_TREE_HH_
#define HATP_PLAN_TREE_HH_

#include<string>
#include<vector>
#include<map>

#include "TaskList.hh"
#include "PlanTreeTaskList.hh"
#include "SocialRuleData.hh"

#define ROOT_LEVEL 0

#define IMMEDIATE_DECOMPOSITION "immediate_decomposition"
#define DEEP_DECOMPOSITION "deep_decomposition"

class Substitution;
class LightWorldBase;

class PlanTree
{
protected:
	/**
	* the root PlanTreeTaskList
	*/
	PlanTreeTaskList* rootList;

	/**
	* the PlanTreeTaskList in the PlanTree
	* (this field is useful to save time during PlanTree exploration)
	*/
	std::vector<PlanTreeTaskList*> PlanTreeNodes;

public:
	explicit PlanTree(TaskList const& TL);
	explicit PlanTree(PlanTree const& PT);
	explicit PlanTree(PlanTreeTaskList* root, std::vector<PlanTreeTaskList*>& vecNodes);
	std::vector<PlanTree*> getMultipleCopies(unsigned long int nbCopies) const;
	std::pair<TaskList*, PlanTree*> makeCopy();
	bool containsNode(unsigned long int argID);
	PlanTreeTaskList* getDecompositionOfID(unsigned long int id) const;
	void decompose(unsigned long int argID, TaskList const& TL);
	std::string toString() const;
	PlanTreeTaskList* getRootList() const;
	std::vector<PlanTreeTaskList*> const& getNodes() const;
	void eraseEmptyBranches();
	void propagateInvolvedAgent();
	void convertToAbstraction();
	unsigned long int getDepth() const;
	std::set<std::string> getTaskNames() const;
	std::vector<PlanTreeTaskListNode*> find(std::string const& taskName) const;
	std::vector<PlanTreeTaskList*> getNodesAtLevel(unsigned long int argLevel) const;
	bool isThereAPathFromTo(unsigned long int IDstart, PlanTreeTaskList* end) const;
	unsigned long int checkSubTree(std::vector<TaskList*> const& subPTnodes,
	std::map<TaskListNode*, std::pair<TaskList*, std::string> > const& hierarchical_links) const;
	std::vector<Substitution*> getSubstitutionsOnNodes(std::vector<TaskList*> const& subPTnodes,
	std::vector<std::string> const& varNames) const;
	std::vector<SocialRuleData*> checkSubTree2(LightWorldBase const& wb,
	std::vector<Substitution*> const& vecSub, std::vector<TaskList*> const& subPTnodes,
	std::map<TaskListNode*, std::pair<TaskList*, std::string> > const& hierarchical_links,
	std::vector<std::string> varNames) const;
	bool operator==(PlanTree const& PT) const;
	PlanTree& operator=(PlanTree const& PT);
	~PlanTree();
	
	bool replaceNode(unsigned long int argID, TaskListNode const& Tln);
};

/**
 * constructor: rootList = new PlanTreeTaskList::PlanTreeTaskList(TL) and add
 * rootList in nodes
 */
inline PlanTree::PlanTree(TaskList const& TL)
	: rootList(new PlanTreeTaskList(TL, ROOT_LEVEL))
{
	PlanTreeNodes.push_back(rootList);
}

/**
* constructor: set rootList to PTTL and add it in nodes
*/
inline PlanTree::PlanTree(PlanTreeTaskList* root, std::vector<PlanTreeTaskList*>& vecNodes)
	: rootList(root), PlanTreeNodes(vecNodes)
{
}

/**
 * return rootList
 */
inline PlanTreeTaskList* PlanTree::getRootList() const
{
	return rootList;
}

/**
 * return nodes
 */
inline std::vector<PlanTreeTaskList*> const& PlanTree::getNodes() const
{
	return PlanTreeNodes;
}

/**
 * return a vector containing all member of PlanTreeNodes which are at Level argLevel
 */
inline std::vector<PlanTreeTaskList*> PlanTree::getNodesAtLevel(unsigned long int argLevel) const
{
	std::vector<PlanTreeTaskList*> vec;

	for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i)
		if(PlanTreeNodes[i]->getLevel() == argLevel)
			vec.push_back(PlanTreeNodes[i]);

	return vec;
}

/**
 * explore PlanTreeNodes and return a set containing all taskName used in the current PlanTree
 */
inline std::set<std::string> PlanTree::getTaskNames() const
{
	std::set<std::string> setExit;
	for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i)
	{
		std::vector<PlanTreeTaskListNode*> const& vec = PlanTreeNodes[i]->getNodes();
		for(unsigned long int j=0 ; j<vec.size() ; ++j)
			setExit.insert(vec[j]->getInstance().getGeneralTask().getName());
	}

	return setExit;
}

/**
 * return true if the two plan trees are the same
 * explore hierarchicaly the Tree comparing nodes at each level
 */
inline bool PlanTree::operator==(PlanTree const& PT) const
{
	return *rootList == *(PT.getRootList());
}

/**
 * destructor: we explore nodes and delete all its members, then clear nodes
 */
inline PlanTree::~PlanTree()
{
	for(unsigned long int i=0 ; i<PlanTreeNodes.size() ; ++i)
		delete PlanTreeNodes[i];
}

#endif
