/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
	 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<string>
#include<iostream>
#include<fstream>
#include<sstream>

#include "PlanTreeTaskList.hh"
#include "tools.hh"

using namespace std;

/**
 * constructor: makes a PlanTreeTaskListNode for each TaskListNode in TL
 * make a copy of causal links in the produces PlanTreeTaskListNodes
 * initializes decompositions to a vector of size = number of nodes in TL
 * and with all its members = NULL
 */
PlanTreeTaskList::PlanTreeTaskList(TaskList const& TL, unsigned long int argLevel)
{
	size_t i,j;
	vector<TaskListNode*> const& vec = TL.getNodes();
	map<TaskListNode*, unsigned long int> AdressToRange;

	nodes.resize(vec.size());
	decompositions = vector<PlanTreeTaskList*>(vec.size(), NULL);
	areLeaves.resize(vec.size());

	for(i=0 ; i<vec.size() ; ++i)
	{
		// we complete nodes
		nodes[i] = new PlanTreeTaskListNode(*vec[i]);
		// we keep correspondance of adress <-> range in nodes vector to make a copy of links
		AdressToRange[vec[i]] = i;

		// we initialize areLeaves
		areLeaves[i] = vec[i]->getInstance().isAction();
	}

	// we make a copy of all causal links of TL
	for(i=0 ; i<vec.size() ; ++i)
	{
		// we make only a copy of predecessors (copy of successors will be done by method
		// AddPredecessors/Successors of class PlanTreeTaskListNode)
		vector<TaskListNode*> const& pred=vec[i]->getPredecessors();
		for(j=0 ; j<pred.size() ; ++j)
			nodes[i]->AddPredecessor(nodes[AdressToRange[pred[j]]]);
	}

	level = argLevel;
}

/**
 * constructor by copying: make a copy of nodes members and update links between nodes
 * according links in PTTL nodes
 * NB: decompositions is duplicated as itself (it must be updated by another function)
 */
PlanTreeTaskList::PlanTreeTaskList(PlanTreeTaskList const& PTTL)
: decompositions(PTTL.getDecompositions()), areLeaves(PTTL.getAreLeaves()), level(PTTL.getLevel())
{
	vector<PlanTreeTaskListNode*> const& PTTLnodes = PTTL.getNodes();
	map<PlanTreeTaskListNode*, unsigned long int> AdressToRange;

	nodes.resize(PTTLnodes.size());

	size_t i,j;
	for(i=0 ; i<PTTLnodes.size() ; ++i)
	{
		// we make a copy of the nodes
		nodes[i] = new PlanTreeTaskListNode(*(PTTLnodes[i]));
		// we keep correspondance of adress <-> range in nodes vector to make a copy of links
		AdressToRange[PTTLnodes[i]] = i;
	}

	// we make a copy of all causal links of PTTL
	for(i=0 ; i<PTTLnodes.size() ; ++i)
	{
		// we make only a copy of predecessors (copy of successors will be done by method
		// AddPredecessors/Successors of class PlanTreeTaskListNode)
		vector<PlanTreeTaskListNode*> const& tmp=PTTLnodes[i]->getPredecessors();
		for(j=0 ; j<tmp.size() ; ++j)
			nodes[i]->AddPredecessor(nodes[AdressToRange[tmp[j]]]);
	}
}

/**
 * explores nodes and return the node with the ID equals to argID else return NULL;
 */
PlanTreeTaskListNode* PlanTreeTaskList::find(unsigned long int argID) const
{
	PlanTreeTaskListNode* p = NULL;
	for(size_t i=0 ; i<nodes.size() ; ++i)
		if(nodes[i]->getID()==argID)
	{
		p = nodes[i];
		break;
	}

	return p;
}

/**
 * add PTTL to the corresponding place in decompositions
 */
void PlanTreeTaskList::decompose(unsigned long int argID, PlanTreeTaskList* PTTL)
{
	size_t i;
	for(i=0 ; i<nodes.size() ; ++i) {
		if(nodes[i]->getID() == argID) {// we have found the node
			if(areLeaves[i])
			{
				cerr << "ERROR: trying to add a decomposition to a PlanTreeTaskList (the node with ID " << argID;
				cerr << " is a leaf so it does not need a decomposition)" << endl;
			}
			else if(decompositions[i] != NULL)
			{
				cerr << "ERROR: trying to add a decomposition to a PlanTreeTaskList (the node with ID " << argID;
				cerr << " have already a decomposition)" << endl;
			}
			else
				decompositions[i] = PTTL;
			break;
		}
	}

	if(i==nodes.size())
	{
		cerr << "ERROR: trying to add a decomposition to a PlanTreeTaskList (no node with ID " << argID;
		cerr << " have been found)" << endl;
	}
}

/**
 * Replace an existing node with a new one
 */
bool PlanTreeTaskList::replaceNode(unsigned long int argID, TaskListNode const& Tln){
	size_t i;
	for(i=0 ; i<nodes.size() ; ++i) {
		if(nodes[i]->getID() == argID) {
			if(!areLeaves[i])
			{
				cerr << "ERROR: trying to replace a node that is not an action (node with ID " << argID << ")" << endl;
				return false;
			}
			if(decompositions[i] != NULL)
			{
				cerr << "ERROR: trying to replace a node that is a method (node with ID " << argID << ")" << endl;
				return false;
			}
			
			//Remove old version and replace withthe new one
			PlanTreeTaskListNode* oldNode=nodes[i];
			nodes[i]=new PlanTreeTaskListNode(Tln);
			
			//Update the information of the node to be the same as the old
			std::vector<PlanTreeTaskListNode*>::const_iterator it;
			std::vector<PlanTreeTaskListNode*> copy(oldNode->getPredecessors()); //Copy because the original will be modified
			for(it=copy.begin(); it!=copy.end(); ++it){
				nodes[i]->AddPredecessor(*it);
				(*it)->RemoveSuccessor(oldNode);
			}
			
			copy=oldNode->getSuccessors();
			for(it=copy.begin(); it!=copy.end(); ++it){
				nodes[i]->AddSuccessor(*it);
				(*it)->RemovePredecessor(oldNode);
			}
			
			nodes[i]->setInvolvedAgents(oldNode->getInvolvedAgents());
			
			//Once done we can remove the old
			delete oldNode;
			break;
		}
	}
	return true;
}
/**
 * converts the PlanTreeTaskList in a string
 * and returns it
 */
string PlanTreeTaskList::toString() const
{
	using namespace HATP_TOOLS;
	size_t i,j;
	ostringstream st;
	string tab;
	for(i=0 ; i<level ; ++i) // to have a lisible printing tabs number is proportional to level
		tab += "  ";
	st << tab << "***** level " << level << " *****\n";
	if(nodes.size()==0)
		st << tab << "()" << endl;
	else {
		st << tab << "-----Nodes-----\n";
		for(i=0 ; i<nodes.size() ; ++i) {
			st << tab << "node " << nodes[i]->getID() << ": " << nodes[i]->getText();
			st << ", agents = " << vectorToString(nodes[i]->getInvolvedAgents()) << endl;
			st << tab << "\tPred: ";
			std::vector<PlanTreeTaskListNode*>::const_iterator it;
			for(it=nodes[i]->getPredecessors().begin(); it!=nodes[i]->getPredecessors().end(); ++it){
				st << (*it)->getID() << ", ";
			}
			st << "\tSucc :";
			for(it=nodes[i]->getSuccessors().begin(); it!=nodes[i]->getSuccessors().end(); ++it){
				st << (*it)->getID() << ", ";
			}
			st << endl;
		}
		st << tab << "-----Links-----\n";
		vector<PlanTreeTaskListNode*> vec;
		for(i=0 ; i<nodes.size() ; ++i) {
			vec=nodes[i]->getSuccessors();
			for(j=0 ; j<vec.size() ; ++j)
				st << tab << "causal link: from " << nodes[i]->getID() << " to " << vec[j]->getID() << endl;
		}
	}
	st << tab << "***** end *****" << endl;

	for(i=0 ; i<decompositions.size() ; ++i)
		if( !(areLeaves[i]) ) {
			if(decompositions[i]==NULL)
				st << tab << "  ***** decomposition of node " << nodes[i]->getID() << " *****\n" << tab << "  private\n" << tab << "  ***** end *****\n";
			else
				st << tab << "  ***** decomposition of node " << nodes[i]->getID() << " *****\n" << decompositions[i]->toString() << tab << "  ***** end *****\n";
		}

	string s(st.str());
	return s;
}

/**
 * explore decompositions and for each empty child (i.e. child with no nodes), we we remove corresponding node in nodes
 * NB: decompositions is not deleted this must be done by another function
 */
void PlanTreeTaskList::eraseEmptyBranches()
{
	vector<PlanTreeTaskListNode*> new_nodes;
	vector<PlanTreeTaskList*> new_decompositions;
	vector<bool> new_areLeaves;
	vector<PlanTreeTaskListNode*> pred, succ;
	size_t i,j,k;
	// we call this function for each existing child
	for(i=0 ; i<decompositions.size() ; ++i)
		if( !areLeaves[i] && decompositions[i]!= NULL)
		decompositions[i]->eraseEmptyBranches();
	for(i=0 ; i<nodes.size() ; ++i) {
		if( !(areLeaves[i]) && decompositions[i] != NULL && decompositions[i]->isEmpty() ) {
			// we update links
			pred = nodes[i]->getPredecessors();
			succ = nodes[i]->getSuccessors();
			for(j=0 ; j<pred.size() ; ++j) // all its predecessors becomes successors of its predecessors
				for(k=0 ; k<succ.size() ; ++k)
				succ[k]->AddPredecessor(pred[j]);
			// we erase all links between nodes[i] and the other nodes
			for(k=0 ; k<pred.size() ; ++k)
				pred[k]->RemoveSuccessor(nodes[i]);
			for(k=0 ; k<succ.size() ; ++k)
				succ[k]->RemovePredecessor(nodes[i]);
			// we free memory
			delete nodes[i];
		}
		else {
			new_nodes.push_back(nodes[i]);
			new_areLeaves.push_back(areLeaves[i]);
			new_decompositions.push_back(decompositions[i]);
		}
	}
	nodes.clear();
	nodes = new_nodes;
	areLeaves.clear();
	areLeaves = new_areLeaves;
	decompositions.clear();
	decompositions = new_decompositions;
}

/**
 * propagates involved agents from the leaves to all the tree
 */
void PlanTreeTaskList::propagateInvolvedAgents()
{
	size_t i;
	for(i=0 ; i<decompositions.size() ; ++i) // we call this function for all children
		if( !areLeaves[i] && decompositions[i] != NULL )
		decompositions[i]->propagateInvolvedAgents();
	for(i=0 ; i<nodes.size() ; ++i) // we merge all vectors from decompositions
		if( !areLeaves[i] && decompositions[i] != NULL )
		nodes[i]->setInvolvedAgents(decompositions[i]->getInvolvedAgents());
}

/**
 * explore nodes and merge vectros of involved agents
 */
vector<string> PlanTreeTaskList::getInvolvedAgents() const
{
	using namespace HATP_TOOLS;
	vector<string> vecAgents;
	for(size_t i=0 ; i<nodes.size() ; ++i)
		vecAgents = vectorMerging(vecAgents, nodes[i]->getInvolvedAgents());
	return vecAgents;
}

/**
 * cut and delete decompositions associating to private task (i.e. task with only one involved agent)
 */
void PlanTreeTaskList::cutPrivateBranches()
{
	size_t i;
	// we call the cutPrivateBranches function for each available decomposition
	for(i=0 ; i<decompositions.size() ; ++i)
		if( !areLeaves[i] && decompositions[i] != NULL)
		decompositions[i]->cutPrivateBranches();
	for(i=0 ; i<nodes.size() ; ++i) // we cut all branches of private nodes
		if( !areLeaves[i] && nodes[i]->getNbInvolvedAgents()==1 && decompositions[i]!=NULL ) {
			delete decompositions[i];
			decompositions[i] = NULL;
		}
}

/**
 * add the current PlanTreeTaskList to tasklists and then call getNodesInTaskListForms(tasklists) for
 * each nodes which are not leaves and have a decomposition.
 * TaskLists in tasklists are sorted by their father IDs, fatherID gives the ID of
 * the current father PlanTreeTaskListNode, oldAndNewIDs allows to keep correspondance between old and new IDs
 */
void PlanTreeTaskList::getNodesInTaskListForms(unsigned long int fatherID, map<unsigned long int, TaskList*>& tasklists, std::map<unsigned long int, unsigned long int>& oldAndNewIDs)
{
	tasklists[fatherID] = new TaskList(this, oldAndNewIDs);
	for(size_t i=0 ; i<nodes.size() ; ++i)
		if( !areLeaves[i] && decompositions[i] != NULL)
		decompositions[i]->getNodesInTaskListForms(nodes[i]->getID(), tasklists, oldAndNewIDs);
}

/**
 * return true if a path exists from IDstart until IDend
 */
bool PlanTreeTaskList::isThereAPathFromTo(unsigned long int IDstart, unsigned long int IDend) const
{
	bool modif;
	size_t i,j;
	vector<PlanTreeTaskListNode*> succ;
	map<PlanTreeTaskListNode*, unsigned long int> PTTLNordersInVector;
	PlanTreeTaskListNode *TLstart=this->find(IDstart), *TLend=this->find(IDend);
	if( TLstart == NULL || TLend == NULL )
		cerr << "ERROR: trying to find a path between two nodes in a PlanTreeTaskList, at least one of these nodes is NULL" << endl;
	else {
		// we create a map associating TaskListNode* and their order in the vectors
		for(i=0 ; i<nodes.size() ; ++i)
			PTTLNordersInVector[nodes[i]] = i;
		// we init connected
		vector<bool> connected(nodes.size(), false);
		connected[PTTLNordersInVector[TLstart]] = true;
		do {
			modif = false;
			for(i=0 ; i<nodes.size() ; ++i)
				if(connected[PTTLNordersInVector[nodes[i]]]) {// if the node is connected all its successors are also connected
					succ = nodes[i]->getSuccessors();
					for(j=0 ; j<succ.size() ; ++j)
						if( !connected[PTTLNordersInVector[succ[j]]] ) {
							connected[PTTLNordersInVector[succ[j]]] = true;
							modif = true;
						}
				}
			if(connected[PTTLNordersInVector[TLend]])
				return true;
		} while(modif);
	}
	// propagation in connected achieved and TLend is not connected
	return false;
}

/**
 * Find all correspondences in the current PlanTreeTaskList from acoording to TaskInstance in subTL
 * and links between them
 */
vector<map<TaskListNode*, PlanTreeTaskListNode*> > PlanTreeTaskList::getSubListCorrespondence(TaskList const& subTL) const {
	size_t i,j,k;
	vector<TaskListNode*> subTLnodes = subTL.getNodes();
	map<TaskListNode*, vector<PlanTreeTaskListNode*> > subTLcandidatesInPTTL;
	// we get all candidates in TL for subTL nodes
	for(i=0 ; i<subTLnodes.size() ; ++i) {
		subTLcandidatesInPTTL[subTLnodes[i]] = vector<PlanTreeTaskListNode*>(0);
		// we explore nodes to find a node with same taskInstance as subTLnodes[i]
		for(j=0 ; j<nodes.size() ; ++j)
			if( subTLnodes[i]->getInstance() == nodes[j]->getInstance() )
			subTLcandidatesInPTTL[subTLnodes[i]].push_back(nodes[j]);
	}
	// we compute all possible combinations by taking two by two
	// the vector members of subTLcandidatesInTL
	vector< map<TaskListNode*, PlanTreeTaskListNode*> > combinations(0), tmp_combinations;
	map<TaskListNode*, PlanTreeTaskListNode*> current_combination;
	map<TaskListNode*, vector<PlanTreeTaskListNode*> >::iterator it;
	for(it = subTLcandidatesInPTTL.begin() ; it != subTLcandidatesInPTTL.end() ; ++it) {
		// initialization
		if(it == subTLcandidatesInPTTL.begin())
			for(i=0 ; i<(it->second).size() ; ++i) {
				combinations.push_back(map<TaskListNode*, PlanTreeTaskListNode*>());
				combinations[i][it->first] = (it->second)[i];
			}
		else {
			// we update compute tmp_combinations from combinations and it->second
			tmp_combinations.clear();
			for(i=0 ; i<(it->second).size() ; ++i)
				for(j=0 ; j<combinations.size() ; ++j) {
					current_combination = combinations[j];
					current_combination[it->first] = (it->second)[i];
					tmp_combinations.push_back(current_combination);
				}
			combinations = tmp_combinations;
		}
	}
	vector<map<TaskListNode*, PlanTreeTaskListNode*> > goodCombinations;
	// for all members in combinations we check the links
	bool good_combination;
	vector<TaskListNode*> succ;
	vector<string> succLinks;
	for(i=0 ; i<combinations.size() ; ++i) {
		good_combination = true;
		for(j=0 ; j<subTLnodes.size() && good_combination ; ++j) {
			succ = subTLnodes[j]->getSuccessors();
			succLinks = subTLnodes[j]->getSuccessorsLinkTypes();
			for(k=0 ; k<succ.size() ; ++k) {
				if( succLinks[k] == IMMEDIATE_LINK ) { // we check if there is a direct link between TaskList nodes
					if( !(combinations[i][subTLnodes[j]]->isAPredecessorOf(combinations[i][succ[k]])) ) {
						good_combination = false;
						break;
					}
				}
				else if( succLinks[k] == LATER_LINK ) {
					// we check if there is a path in the TaskList from subTLnodes[i] to succ[j]
					if( !(this->isThereAPathFromTo(combinations[i][subTLnodes[j]]->getID(), combinations[i][succ[k]]->getID())) ) {
						good_combination = false;
						break;
					}
				}
				else {// error link type is not recognized
					cerr << "ERROR when checking if a TaskList contains a subTaskList, link type \"" << succLinks[k];
					cerr << "\" unrecognized" << endl;
					return vector<map<TaskListNode*, PlanTreeTaskListNode*> >(0);
				}
			}
		}
		if(good_combination)
			goodCombinations.push_back(combinations[i]);
	}
	return goodCombinations;
}

/**
 * returns true if the PlanTreeTaskList is equals to t (this operator does not
 * consider IDs to compare the PlanTreeTaskLists)
 */
bool PlanTreeTaskList::operator==(PlanTreeTaskList const& t) const
{
	size_t i,j;
	vector<PlanTreeTaskListNode*> Tnodes = t.getNodes();
	vector<bool> Tleaves = t.getAreLeaves();
	vector<PlanTreeTaskList*> Tdecompo = t.getDecompositions();
	string sNodes_i;

	if(level == t.getLevel() && nodes.size() == Tnodes.size())
	{
		for(i=0 ; i<nodes.size() ; ++i)
		{
			sNodes_i = nodes[i]->getText();

			// we have to find a node with same task, decompositions, predecessors and successors
			for(j=0 ; j<Tnodes.size() ; ++j)
				if( sNodes_i == Tnodes[j]->getText() && nodes[i]->hasSamePredecessors(Tnodes[j]) &&
				   nodes[i]->hasSameSuccessors(Tnodes[j]) && areLeaves[i]==Tleaves[j] &&
				   ( areLeaves[i] || ( !areLeaves[i] && *(decompositions[i]) == *(Tdecompo[j]) ) ) )
				break;

			if(j == Tnodes.size())
				return false;
		}

		return true;
	}

	return false;
}

/**
 * assignment operator (same as a copy of PTTL)
 */
PlanTreeTaskList& PlanTreeTaskList::operator=(PlanTreeTaskList const& PTTL)
{
	size_t i;

	// we clear nodes, decompositions and areLeaves
	for(i=0 ; i<nodes.size() ; ++i)
		delete nodes[i];
	nodes.clear();
	decompositions.clear();
	areLeaves.clear();

	// we update data
	vector<PlanTreeTaskListNode*> const& PTTLnodes = PTTL.getNodes();
	map<PlanTreeTaskListNode*, unsigned long int> AdressToRange;

	nodes.resize(PTTLnodes.size());

	size_t j;
	for(i=0 ; i<PTTLnodes.size() ; ++i)
	{
		// we make a copy of the nodes
		nodes[i] = new PlanTreeTaskListNode(*(PTTLnodes[i]));
		// we keep correspondance of adress <-> range in nodes vector to make a copy of links
		AdressToRange[PTTLnodes[i]] = i;
	}

	// we make a copy of all causal links of PTTL
	for(i=0 ; i<PTTLnodes.size() ; ++i)
	{
		// we make only a copy of predecessors (copy of successors will be done by method
		// AddPredecessors/Successors of class PlanTreeTaskListNode)
		vector<PlanTreeTaskListNode*> const& tmp=PTTLnodes[i]->getPredecessors();
		for(j=0 ; j<tmp.size() ; ++j)
			nodes[i]->AddPredecessor(nodes[AdressToRange[tmp[j]]]);
	}

	decompositions = PTTL.getDecompositions();
	areLeaves = PTTL.getAreLeaves();
	level = PTTL.getLevel();

	return *this;
}

/**
 * get all nodes of the tree starting from this PlanTreeTaskList
 */
vector<PlanTreeTaskListNode*> PlanTreeTaskList::getAllNodes() const {
	vector<PlanTreeTaskListNode*> current = this->getNodes();
	for(size_t i = 0 ; i < decompositions.size() ; i++) {
		if(decompositions[i] != NULL) {
			vector<PlanTreeTaskListNode*> subnodes = decompositions[i]->getAllNodes();
			for(size_t j = 0 ; j < subnodes.size() ; j++) current.push_back(subnodes[j]);
		}
	}
	return current;
}

vector<pair<unsigned int, unsigned int> > PlanTreeTaskList::getHierarchicalLinks() {
	vector<pair<unsigned int, unsigned int> > result;
	vector<PlanTreeTaskListNode*> cnodes = this->getNodes();
	for(size_t i = 0 ; i < cnodes.size() ; i++) {
		PlanTreeTaskList* curDecompo = this->getDecompositionOfID((cnodes[i])->getID());
		if(curDecompo != NULL) {
			vector<PlanTreeTaskListNode*> hnodes = curDecompo->getNodes();
			for(size_t j = 0 ; j < hnodes.size() ; j++) {
				pair<unsigned int, unsigned int> hltemp = make_pair(cnodes[i]->getID(), hnodes[j]->getID());
				result.push_back(hltemp);
			}
		}
	}
	for(size_t i = 0 ; i < decompositions.size() ; i++) {
		if(decompositions[i] != NULL) {
			vector<pair<unsigned int, unsigned int> > others = decompositions[i]->getHierarchicalLinks();
			for(size_t j = 0 ; j < others.size() ; j++) result.push_back(others[j]);
		}
	}
	return result;
}





