/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_PLAN_TREE_TASK_LIST_HH_
#define HATP_PLAN_TREE_TASK_LIST_HH_

#include<string>
#include<vector>
#include<map>
#include<iostream>
#include<sstream>

#include "TaskInstance.hh"
#include "TaskListNode.hh"
#include "TaskList.hh"
#include "PlanTreeTaskListNode.hh"

class PlanTreeTaskList
{
protected:
	/**
	* the list of PlanTreeTaskListNodes in the PlanTreeTaskList
	*/
	std::vector<PlanTreeTaskListNode*> nodes;

	/**
	* the decompositions of all members of nodes
	* arrange in the same order than nodes
	*/
	std::vector<PlanTreeTaskList*> decompositions;

	/**
	* true if the associated decomposition is a leaf for the tree
	* arrange in the same order than nodes
	*/
	std::vector<bool> areLeaves;

	/**
	* depth level of the PlanTreeTaskList
	*/
	unsigned long int level;

public:
	explicit PlanTreeTaskList(TaskList const& TL, unsigned long int argLevel);
	explicit PlanTreeTaskList(PlanTreeTaskList const& PTTL);
	explicit PlanTreeTaskList(std::vector<PlanTreeTaskListNode*> const& argNodes,
		std::vector<PlanTreeTaskList*> const& argDecompositions, std::vector<bool> const& argAreLeaves,
		unsigned long int argLevel);
	PlanTreeTaskListNode* find(unsigned long int argID) const;
	bool containsNode(unsigned long int argID) const;
	PlanTreeTaskList* getDecompositionOfID(unsigned long int argID);
	bool isEmpty() const;
	unsigned long int getLevel() const;
	void decompose(unsigned long int argID, PlanTreeTaskList* PTTL);
	std::string toString() const;
	std::vector<PlanTreeTaskListNode*> const& getNodes() const;
	std::vector<PlanTreeTaskList*> const& getDecompositions() const;
	void setDecompositions(std::vector<PlanTreeTaskList*> const& decompos);
	std::vector<bool> const& getAreLeaves() const;
	void eraseEmptyBranches();
	unsigned long int getNbNodes() const;
	void propagateInvolvedAgents();
	std::vector<std::string> getInvolvedAgents() const;
	unsigned long int getNbInvolvedAgents() const;
	void cutPrivateBranches();
	void getNodesInTaskListForms(unsigned long int fatherID,
		std::map<unsigned long int, TaskList*>& tasklists,
		std::map<unsigned long int, unsigned long int>& oldAndNewIDs);
	bool isThereAPathFromTo(unsigned long int IDstart, unsigned long int IDend) const;
	std::vector< std::map<TaskListNode*, PlanTreeTaskListNode*> >
		getSubListCorrespondence(TaskList const& subTL) const;
	bool operator==(PlanTreeTaskList const& t) const;
	PlanTreeTaskList& operator=(PlanTreeTaskList const& PTTL);
	~PlanTreeTaskList();
	std::vector<PlanTreeTaskListNode*> getAllNodes() const;
	std::vector<std::pair<unsigned int, unsigned int> > getHierarchicalLinks();
	
	bool replaceNode(unsigned long int argID, TaskListNode const& Tln);
};


/**
* constructor: set nodes to argNodes, decompositions to argDecompositions and areLeaves to argAreLeaves
*/
inline PlanTreeTaskList::PlanTreeTaskList(std::vector<PlanTreeTaskListNode*> const& argNodes,
	std::vector<PlanTreeTaskList*> const& argDecompositions, std::vector<bool> const& argAreLeaves,
	unsigned long int argLevel)
	: nodes(argNodes), decompositions(argDecompositions), areLeaves(argAreLeaves), level(argLevel)
{
}

/**
 * return true if the current PlanTreeTaskList contains a PlanTreeTaskListNode with ID = argID
 */
inline bool PlanTreeTaskList::containsNode(unsigned long int argID) const
{
	PlanTreeTaskListNode* p = this->find(argID);
	if(p != NULL) // the task with ID = argID has been found at this level
		return true;

	return false;
}

/**
 * return a pointer on decomposition associated to node with ID id, if the node with
 * ID id is not found return NULL
 */
inline PlanTreeTaskList* PlanTreeTaskList::getDecompositionOfID(unsigned long int argID)
{
	for(unsigned long int i=0 ; i<nodes.size() ; ++i)
		if(nodes[i]->getID() == argID)
			return decompositions[i];

	return NULL;
}

/**
 * return true if nodes is empty
 */
inline bool PlanTreeTaskList::isEmpty() const
{
	return nodes.empty();
}

/**
 * return level
 */
inline unsigned long int PlanTreeTaskList::getLevel() const
{
	return level;
}

/**
 * return nodes
 */
inline std::vector<PlanTreeTaskListNode*> const& PlanTreeTaskList::getNodes() const
{
	return nodes;
}

/**
 * return decompositions
 */
inline std::vector<PlanTreeTaskList*> const& PlanTreeTaskList::getDecompositions() const
{
	return decompositions;
}

/**
 * set decompositions to decompos
 */
inline void PlanTreeTaskList::setDecompositions(std::vector<PlanTreeTaskList*> const& decompos)
{
	if(decompos.size() == decompositions.size())
		decompositions = decompos;
	else
	{
		std::cerr << "ERROR when trying to affect decompositions of PlanTreeTaskList:";
		std::cerr << " the input vector has not the good size." << std::endl;
	}
}

/**
 * return areLeaves
 */
inline std::vector<bool> const& PlanTreeTaskList::getAreLeaves() const
{
	return areLeaves;
}

/**
 * return size of nodes
 */
inline unsigned long int PlanTreeTaskList::getNbNodes() const
{
	return nodes.size();
}

/**
 * call this->getInvolvedAgents() and return the size of the resulting vector
 */
inline unsigned long int PlanTreeTaskList::getNbInvolvedAgents() const
{
	return (this->getInvolvedAgents()).size();
}

/**
 * destructor: free memory by deleting all members of nodes
 */
inline PlanTreeTaskList::~PlanTreeTaskList()
{
	for(unsigned long int i=0 ; i<nodes.size() ; ++i)
		delete nodes[i];
}

#endif
