#include<iostream>

#include "TaskInstance.hh"
#include "WorldBase.hh"
#include "LightWorldBase.hh"
#include "HTNBase.hh"

using namespace std;

HTNBase* TaskInstance::htn_base = NULL;

/**
 * constructor: get generalTask fron htn_base and taskName, set vecArgs to args and
 * call generalTask->checkTypes.
 */
TaskInstance::TaskInstance(string const& taskName, vector<string> const& vecArgs, LightWorldBase const& wb)
: generalTask(htn_base->findConst(taskName)), uniqueId(-1), args(vecArgs), originalInstance(NULL) {
	if( !(generalTask.checkTypes(args, wb)) )
		cerr << "ERROR when creating a TaskInstance for Task " << taskName << ", bad argument types." << endl;
}

/**
 * constructor: get generalTask fron htn_base and taskName, set vecArgs to args and
 * call generalTask->checkTypes with wb.
 */
TaskInstance::TaskInstance(string const& taskName, vector<string> const& vecArgs, WorldBase const& wb)
: generalTask(htn_base->findConst(taskName)), uniqueId(-1), args(vecArgs), originalInstance(NULL) {
	if( !(generalTask.checkTypes(args, wb)) )
		cerr << "ERROR when creating a TaskInstance for Task " << taskName << ", bad argument types." << endl;
}

/**
 * constructor: get generalTask fron htn_base and taskName, set vecArgs to args and call generalTask->checkTypes with varAndTypes.
 * This constructor is useful for partial TaskList/PlanTree for undesirable sequences and bad decompositions rules.
 */
TaskInstance::TaskInstance(string const& taskName, vector<string> const& vecArgs,
 map<string, string> const& varAndTypes) : generalTask(htn_base->findConst(taskName)), uniqueId(-1),
 	args(vecArgs), originalInstance(NULL) {
	if( !(generalTask.checkTypes(args, varAndTypes)) ) {
		cerr << "ERROR when creating a TaskInstance for Task " << taskName;
		cerr << " with external variables, bad argument types." << endl;
	}
}

/**
 * if argHTNBase is not NULL, delete htn_base and set it to argHTNBase
 */
void TaskInstance::setHTNBase(HTNBase* argHTNBase) {
	if(argHTNBase) {
		delete htn_base;
		htn_base = argHTNBase;
	}
}

/**
 * return result of checkPreconds(args)
 */
bool TaskInstance::checkPreconditions(LightWorldBase& wb, const hatpPlanningSession& session) {
	unsigned int nbArgs = generalTask.getNbArgs();
	vector<LightEntity const*> vec(nbArgs);
	precondWithReportFunc fct = generalTask.getCheckingPreconditionsWithReportFunction();
	if(fct==NULL) return true;
	if(!generalTask.isCommunicationTask()) {
		for(unsigned int i=0 ; i<nbArgs ; ++i) vec[i] = wb.find(args[i]);
	}
	else {
		for(unsigned int i=0 ; i<nbArgs-1 ; ++i) vec[i] = wb.find(args[i]);
		vec[nbArgs-1] = new LightEntity(args[nbArgs-1]);
	}
	return fct(wb, vec, *this, precondReport, session);
}

/**
 * return the report from the preconditions (computed in checkPreconditions)
 */
void TaskInstance::getPreconditionsReport(WorldBaseReport& report) const{
	report=precondReport;
}

/**
 * return result of checkEmpty(args)
 */
bool TaskInstance::checkTaskEmpty(LightWorldBase const& wb, hatpPlanningSession const& session) const {
	unsigned int nbArgs = generalTask.getNbArgs();
	vector<LightEntity const*> vec(nbArgs);
	emptyFunc fct = generalTask.getCheckingEmptyFunction();
	if(fct==NULL) // if no function to check empty, it means the task is not empty
		return false;
	for(unsigned int i=0 ; i<nbArgs ; ++i)
		vec[i] = wb.find(args[i]);
	return fct(wb, vec, *this, session);
}

/**
 * returns result of getDecompo(args)
 */
vector<TaskList*> TaskInstance::getDecompositions(LightWorldBase const& wb, HTNBase const& htnb, hatpPlanningSession const& session) const {
	unsigned int nbArgs = generalTask.getNbArgs();
	vector<LightEntity const*> vec(nbArgs);
	decompoFunc fct = generalTask.getDecompositionsFunction();
	if(fct==NULL) // if no function to get decompositions, it means no decompositions available
		return vector<TaskList*>(0);
	for(unsigned int i=0 ; i<nbArgs ; ++i)
		vec[i] = wb.find(args[i]);
	return fct(wb, htnb, vec, *this, session);
}

/**
 * calls applyEffects
 */
void TaskInstance::applyActionEffects(LightWorldBase& wb) const {
	effectsFunc fct = generalTask.getApplyEffectsFunction();
	if(fct != NULL) { // if no function to apply effects, it means no effects to apply
		if(generalTask.isAction()) {
			unsigned int nbArgs = generalTask.getNbArgs();
			vector<LightEntity*> vec(nbArgs);
			if(!generalTask.isCommunicationTask()) {
				for(unsigned int i=0 ; i<nbArgs ; ++i) vec[i] = wb.find(args[i]);
			}
			else {
				for(unsigned int i=0 ; i<nbArgs-1 ; ++i) vec[i] = wb.find(args[i]);
				vec[nbArgs-1] = new LightEntity(args[nbArgs-1]);
			}
			fct(wb, vec, *this);	
		}
		else cerr << "ERROR when trying to applyEffects of a Task, the task is not an action" << endl;
	}
}

/**
 * calls applyEffectsWithReport
 */
void TaskInstance::applyActionEffectsWithReport(LightWorldBase& wb, WorldBaseReport& report) const {
	effectsWithReportFunc fct = generalTask.getApplyEffectsWithReportFunction();
	if(fct != NULL) { // if no function to apply effects, it means no effects to apply
		if(generalTask.isAction()) {
			unsigned int nbArgs = generalTask.getNbArgs();
			vector<LightEntity*> vec(nbArgs);
			for(unsigned int i=0 ; i<nbArgs ; ++i)
				vec[i] = wb.find(args[i]);
			fct(wb, vec, *this, report);
		}
		else cerr << "ERROR when trying to applyEffectsWithReports with report of a Task, the task is not an action" << endl;
	}
}

/**
 * calls computeCost and return the result
 */
float TaskInstance::ComputeCost(LightWorldBase const& wb) const {
	costFunc fct = generalTask.getComputeCostFunction();
	if(fct != NULL) { // if no function to compute cost, it means it is equals to 0
		if(generalTask.isAction()) {
			unsigned int nbArgs = generalTask.getNbArgs();
			vector<LightEntity const*> vec(nbArgs);
			if(!generalTask.isCommunicationTask()) {
				for(unsigned int i=0 ; i<nbArgs ; ++i) vec[i] = wb.find(args[i]);
			}
			else {
				for(unsigned int i=0 ; i<nbArgs-1 ; ++i) vec[i] = wb.find(args[i]);
				vec[nbArgs-1] = new LightEntity(args[nbArgs-1]);
			}
			return fct(wb, vec, *this);
				
		}
		else cerr << "ERROR when trying to compute cost of a Task, the task is not an action" << endl;
	}
	return 0;
}

/**
 * calls computeTimeLimit and return the result
 */
void TaskInstance::ComputeTimeLimit(LightWorldBase const& wb) {
	timeFunc fct = generalTask.getComputeTimeFunction();
	if(fct != NULL) { // if no function to compute time, it means it is equals to 1
		if(generalTask.isAction()) {
			unsigned int nbArgs = generalTask.getNbArgs();
			vector<LightEntity const*> vec(nbArgs);
			if(!generalTask.isCommunicationTask()) {
				for(unsigned int i=0 ; i<nbArgs ; ++i) vec[i] = wb.find(args[i]);
			}
			else {
				for(unsigned int i=0 ; i<nbArgs-1 ; ++i) vec[i] = wb.find(args[i]);
				vec[nbArgs-1] = new LightEntity(args[nbArgs-1]);
			}
			pair<double,double> timeLimits=fct(wb, vec);
			minTime = timeLimits.first;
			maxTime = timeLimits.second;
		}
		else cerr << "ERROR when trying to compute time of a Task, the task is not an action" << endl;
	}
	else {minTime = 1.; maxTime = 1.;}
}

/**
 * produce a string from the generalTask name and args and return it: name(name_arg1, name_arg2,....)
 */
string TaskInstance::toString() const {
	stringstream s;
	s << generalTask.getName() << "(";
	for(unsigned int i=0 ; i<args.size() ; ++i) {
		s << args[i];
		if(i<args.size()-1) s << ", ";
	}
	s << "):" << uniqueId;
	return s.str();
}

/**
 * if generalTask is an action, return the list of agents associated to the
 * task (i.e. the list of first parameters which are of type AGENT)
 * else prints an error and return an empty vector
 */
vector<string> TaskInstance::getAssociatedAgents() const {
	if( !(generalTask.isAction()) ) {
		cerr << "ERROR: trying to get associated Agents of the Task ";
		cerr << this->toString() << " which is not an action (return an empty vector)" << endl;
		return vector<string>(0);
	}
	else {
		unsigned int nbAgents = generalTask.getNbInvolvedAgents();
		return vector<string>(args.begin(), args.begin() + nbAgents);
	}
}

/**
 * return true if the two instances have the same uniqueId
 * return true if current TaskInstance and ti have the same generalTask and same args
 */
bool TaskInstance::operator==(TaskInstance const& ti) const {
	if(this->uniqueId != -1 && ti.uniqueId != -1){
		return (this->uniqueId == ti.uniqueId);
	}
	
	//Else return true if current TaskInstance and ti have the same generalTask and same args
	vector<string> tiArgs = ti.getArgs();
	if( generalTask.getName() == ti.getGeneralTask().getName() && args.size() == tiArgs.size() ) {
		for(unsigned int i=0 ; i<args.size() ; ++i)
			if(args[i] != tiArgs[i]) return false;
		return true;
	}
	return false;
}

