#if !defined HATP_TASKLIST_HH_
#define HATP_TASKLIST_HH_

#include<string>
#include<vector>
#include<set>
#include<map>

#include "TaskListNode.hh"
#include "SocialRuleData.hh"


class TaskInstance;
class PlanTreeTaskList;
class PlanTreeTaskListNode;
class Substitution;
class LightEntity;
class Entity;
class LightWorldBase;
class WorldBase;
class hatpPlanningSession;

class TaskList
{
protected:
	/**
	* the list of TaskListNodes in the TaskList
	*/
	std::vector<TaskListNode*> nodes;

	/**
	* define if nodes must be deleted
	*/
	bool nodes2BeDeleted;

	bool lastProjectedTaskAssociatedToAgent(std::string const& Ag, unsigned long int& lastTsk) const;

public:
	explicit TaskList();
	explicit TaskList(std::vector<TaskListNode*> const& argNodes);
	explicit TaskList(TaskList const& TL, bool reviseIDs);
	TaskList(TaskList const& TL);
	explicit TaskList(std::map<unsigned int, TaskInstance*> const& tasks, std::map<unsigned int,std::vector<unsigned int> > const& links,
		std::map<unsigned long int, unsigned int>* IDs = NULL);
	explicit TaskList(std::map<unsigned int, TaskInstance*> const& tasks, std::map<unsigned int, std::vector<unsigned int> > const& links,
		std::map<unsigned int, std::vector<std::string> > const& linkTypes, std::map<unsigned long int, unsigned int>* IDs = NULL);
	explicit TaskList(PlanTreeTaskList const* PTTL, std::map<unsigned long int, unsigned long int>& oldAndNewIDs);
	std::vector<TaskList*> getMultipleCopies(unsigned int nbCopies);
	void addNodes(std::vector<TaskListNode*> const& vec);
	TaskListNode* find(unsigned long int argID) const;
	TaskListNode* find(TaskInstance const& t) const;
	std::vector<TaskListNode*> find(std::string const& taskName) const;
	std::vector<TaskListNode*> findNodesWithInstance(TaskInstance const& t) const;
	std::set<std::string> getTaskNames() const;
	std::vector<unsigned long int> getFirstTasks(LightWorldBase& wb, const hatpPlanningSession& session);
	std::vector<unsigned long int> getFirstUnlockedTasks(LightWorldBase& wb, const hatpPlanningSession& session);
	std::vector<TaskListNode*> const& getNodes() const;
	std::vector<TaskListNode*> getCopyOfNodes() const;
	void substitute(unsigned long int argID, TaskList const& TL);
	void substitute(unsigned long int argID, TaskList* TL);
	void setStatus(unsigned long int ID, int status);
	void setAllStatus(int status);
	std::string toString() const;
	bool isEmpty() const;
	bool allDone() const;
	void unlockAll();
	void project(unsigned long int tskToProject, float costTskProjection=0);
	std::map<std::string, float> getEfforts() const;
	std::multimap<std::vector<LightEntity const*>, std::vector<LightEntity const*> > getIntricateLinks(LightWorldBase const& wb) const;
	std::multimap<std::vector<LightEntity const*>, std::vector<LightEntity const*> > getIntricateLinksFromTo(LightWorldBase const& wb, std::vector<std::string>& startSet, bool startStrict, std::vector<std::string>& destSet, bool destStrict) const;
	unsigned int getNbIntricateLinks(std::set<std::string> const& robots) const;
	void destroyAllLinksOfType(const std::string& sLinkType);
	bool isThereAPathFromTo(unsigned int IDstart, unsigned int IDend) const;
	unsigned int checkSubList(TaskList const& subTL, std::map<unsigned long int, std::pair<double, double> > const& ptimes) const;
	std::vector<Substitution*> getSubstitutionsOnNodes(TaskList const& subTL, std::vector<std::string> const& varNames) const;
	TaskList* Rebuild(Substitution const& subst, LightWorldBase const& wb) const;
	TaskList* Rebuild(Substitution const& subst, LightWorldBase const& wb, std::map<TaskListNode*, TaskListNode*>& nodesToExitNodes) const;
	void noNeedToDeleteNodes();
	unsigned int getNbLinks() const;
	std::vector<unsigned long int> getOrderedTaskIDs(std::string const& agentName) const;
	bool operator==(TaskList const& TL) const;
	TaskList& operator=(TaskList const& t);
	~TaskList();
};

/**
 * constructor: empty
 */
inline TaskList::TaskList() : nodes2BeDeleted(true)
{
}

/**
 * constructor: sets nodes to argNodes
 */
inline TaskList::TaskList(std::vector<TaskListNode*> const& argNodes) : nodes(argNodes), nodes2BeDeleted(true)
{
}

/**
 * constructor by copying
 */
inline TaskList::TaskList(TaskList const& TL) : nodes2BeDeleted(true)
{
	// we complete nodes with TL nodes
	std::vector<std::string> tmp_link_types;
	std::vector<TaskListNode*> const& vec=TL.getNodes();
	std::map<TaskListNode*, unsigned int> AdressToRange;
	unsigned int i,j;
	for(i=0 ; i<vec.size() ; ++i) {
		nodes.push_back(new TaskListNode(*vec[i], false));
		// we keep correspondance of adress <-> range in nodes vector to make a copy of links
		AdressToRange[vec[i]] = i;
	}
	// we make a copy of all causal links of TL
	for(i=0 ; i<vec.size() ; ++i) {
		// we make only a copy of predecessors (copy of successors will be done by method AddPredecessors/Successors of class TaskListNode)
		std::vector<TaskListNode*> const& tmp=vec[i]->getPredecessors();
		tmp_link_types=vec[i]->getPredecessorsLinkTypes();
		for(j=0 ; j<tmp.size() ; ++j)
			nodes[i]->AddPredecessor(nodes[AdressToRange[tmp[j]]], tmp_link_types[j]);
	}
}

/**
 * explores vec and add all of its members in nodes
 */
inline void TaskList::addNodes(std::vector<TaskListNode*> const& vec)
{
	unsigned int start = nodes.size();
	nodes.resize(nodes.size() + vec.size());
	for(unsigned int i=0 ; i<vec.size() ; ++i)
		nodes[start+i] = vec[i];
}

/**
 * return nodes
 */
inline std::vector<TaskListNode*> const& TaskList::getNodes() const
{
	return nodes;
}

/**
 * explore nodes and return a set containing all taskName used in the current TaskList
 */
inline std::set<std::string> TaskList::getTaskNames() const
{
	std::set<std::string> setExit;
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		setExit.insert(nodes[i]->getInstance().getGeneralTask().getName());
	return setExit;
}

/**
 * returns true if the TaskList is empty
 */
inline bool TaskList::isEmpty() const
{
	return nodes.empty();
}

/**
 * returns true if all the tasks asociated to nodes have the DONE status
 */
inline bool TaskList::allDone() const
{
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		if(nodes[i]->getStatus() != DONE)
			return false;
	return true;
}

/**
 * explore nodes and for all nodes with the status TODO_LOCKED set their status to TODO
 */
inline void TaskList::unlockAll()
{
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		if(nodes[i]->getStatus() == TODO_LOCKED)
			nodes[i]->setStatus(TODO);
}

/**
 * set nodes2BeDeleted to false
 */
inline void TaskList:: noNeedToDeleteNodes()
{
	nodes2BeDeleted = false;
}

/**
 * return the number of links in the current TaskList
 */
inline unsigned int TaskList::getNbLinks() const
{
	unsigned int nb=0;
	for(unsigned int i=0 ; i<nodes.size() ; ++i)
		nb += nodes[i]->getNbSuccessors();
	return nb;
}

#endif
