/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_CHOICE_POINT_HH_
#define HATP_CHOICE_POINT_HH_

#include<vector>
#include<string>

#include "LightWorldBase.hh"
#include "Task.hh"
#include "PartialPlanAppraiser.hh"
#include "TaskListNode.hh"
#include "TaskList.hh"
#include "PlanTree.hh"

class ChoicePoint
{
public:
	/** Type of the choice point :
	 */
	enum Type {ROOT_TYPE, //The root of ChoiceReminder
		DECOMPO_TYPE, //Planning faced a decomposition choice
		PARALLEL_TYPE}; //Planning faced a parallel choice

	/** Possible states of the choice point :
	 */
	enum Tag {CURRENT_TAG, //Planning's current point
		UNEXPLORED_TAG, ///Choice not explored
		EXPLORED_TAG, //Choice expolored
		SUCCESS_TAG, //Choice gave a possible plan
		FAILED_TAG, //Choice did not allow to reach a plan
		COST_EXCEEDED_TAG}; //Choice current cost is higher than best plan cost, do not explore further
private:
	/** The type of the choice point
	 */
	Type type;

	/** Choice point state
	 */
	Tag tag;

	/**
	* the WorldBase associated to the choice point
	*/
	LightWorldBase* worldbase;

	/**
	* the TaskList associated to the choice point
	*/
	TaskList* tasklist;

	/**
	* the vector defining order in which tasks have been projected
	*/
	std::vector<unsigned long int> projection_order;

	/**
	* the Task (member of tasklist) associated to the choice point
	*/
	unsigned long int task;

	/**
	* the PlanTree associated to the choice point
	*/
	PlanTree* tree;

	/**
	* the cost of thre projection
	*/
	PartialPlanAppraiser* partial_eval;

	/**
	* the children of the choice point
	*/
	std::vector<ChoicePoint*> next;

	/**
	* the father of the choice point
	*/
	ChoicePoint* father;

public:
	explicit ChoicePoint(ChoicePoint::Type sType, LightWorldBase const& wb, TaskList const& TL,
		std::vector<unsigned long int> const& po, PlanTree const& PT, PartialPlanAppraiser const& pEval);
	explicit ChoicePoint(ChoicePoint::Type sType, WorldBase const& wb, TaskList const& TL,
		std::vector<unsigned long int> const& po, PlanTree const& PT, PartialPlanAppraiser const& pEval);
	explicit ChoicePoint(ChoicePoint::Type sType, LightWorldBase const& wb, TaskList* TL,
		std::vector<unsigned long int> const& po, PlanTree* PT, PartialPlanAppraiser const& pEval);
	explicit ChoicePoint(ChoicePoint::Type sType, LightWorldBase const& wb, TaskList const& TL,
		std::vector<unsigned long int> const& po, PlanTree const& PT, unsigned long int TaskID,
		PartialPlanAppraiser const& pEval);
	explicit ChoicePoint(ChoicePoint::Type sType, LightWorldBase const& wb, TaskList* TL,
		std::vector<unsigned long int> const& po, PlanTree* PT, unsigned long int TaskID,
	PartialPlanAppraiser const& pEval);
	explicit ChoicePoint(ChoicePoint const& cp);

	Tag const& getTag() const;
	void setTag(Tag sTag);
	Tag propagateTags();

	void addChild(ChoicePoint* CP);
	void setFather(ChoicePoint* CP);
	ChoicePoint* getFather() const;
	std::vector<ChoicePoint*> const& getChildren() const;
	void deleteChildren();
	bool hasNoCurrentChild() const;

	ChoicePoint::Type getType() const;
	LightWorldBase* getWorldBase();
	LightWorldBase const* getWorldBase() const;
	TaskList* getTaskList();
	TaskList const* getTaskList() const;
	std::vector<unsigned long int> const& getProjectionOrder() const;
	unsigned long int getTask() const;
	PlanTree* getTree();
	PlanTree const* getTree() const;
	PartialPlanAppraiser const* getPartialEval() const;
	PartialPlanAppraiser* getPartialEval();

	std::string toString() const;
	static std::string typeToString(Type type);
	static std::string tagToString(Tag tag);
	~ChoicePoint();
};

/**
 * returns tag
 */
inline ChoicePoint::Tag const& ChoicePoint::getTag() const
{
	return tag;
}

/**
 * returns type
 */
inline ChoicePoint::Type ChoicePoint::getType() const
{
	return type;
}

/** Set the tag to the given one
 * @param sTag The new tag
 */
inline void ChoicePoint::setTag(ChoicePoint::Tag sTag)
{
	tag = sTag;
}

/**
 * adds CP in next, sets CP->father to this
 */
inline void ChoicePoint::addChild(ChoicePoint* CP)
{
	next.push_back(CP);
	CP->setFather(this);
}

/**
 * sets father to CP
 */
inline void ChoicePoint::setFather(ChoicePoint* CP)
{
	father = CP;
}

/**
 * return father
 */
inline ChoicePoint* ChoicePoint::getFather() const
{
	return father;
}

/**
 * return next
 */
inline std::vector<ChoicePoint*> const& ChoicePoint::getChildren() const
{
	return next;
}

/**
 * return worldbase
 */
inline LightWorldBase* ChoicePoint::getWorldBase()
{
	return worldbase;
}

/**
 * return worldbase (const version)
 */
inline LightWorldBase const* ChoicePoint::getWorldBase() const
{
	return worldbase;
}

/**
 * return tasklist
 */
inline TaskList* ChoicePoint::getTaskList()
{
	return tasklist;
}

/**
 * return tasklist (const version)
 */
inline TaskList const* ChoicePoint::getTaskList() const
{
	return tasklist;
}

/**
 * return projection_order
 */
inline std::vector<unsigned long int> const& ChoicePoint::getProjectionOrder() const
{
	return projection_order;
}

/**
 * return task
 */
inline unsigned long int ChoicePoint::getTask() const
{
	return task;
}

/**
 * return tree
 */
inline PlanTree* ChoicePoint::getTree()
{
	return tree;
}

/**
 * return tree (const version)
 */
inline PlanTree const* ChoicePoint::getTree() const
{
	return tree;
}

/**
 * return partial_eval
 */
inline PartialPlanAppraiser* ChoicePoint::getPartialEval()
{
	return partial_eval;
}

/**
 * return partial_eval (const version)
 */
inline PartialPlanAppraiser const* ChoicePoint::getPartialEval() const
{
	return partial_eval;
}

/**
 * destructor: delete worldbase, tasklist, tree and partial_eval
 */
inline ChoicePoint::~ChoicePoint()
{
	delete worldbase;
	delete tasklist;
	delete tree;
	delete partial_eval;
}

#endif
