/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_CHOICE_REMINDER_HH_
#define HATP_CHOICE_REMINDER_HH_

#include<vector>
#include<set>
#include<map>

#include "ChoicePoint.hh"
#include "WorldBase.hh"
#include "LightWorldBase.hh"
#include "PartialPlanAppraiser.hh"
#include "TaskList.hh"
#include "PlanTree.hh"

class ChoiceReminder
{
private:
	/**
	* the root node of the ChoiceReminder
	*/
	ChoicePoint* root_point;

	/**
	* the current node of the ChoiceReminder (it is just to accelerate the research in the ChoiceReminder)
	*/
	ChoicePoint* current_point;

	/**
	* the list of ChoicePoints
	*/
	std::set<ChoicePoint*> points;
	
	std::string recursiveBranchPrint(ChoicePoint* point, std::string offset) const;

public:
	explicit ChoiceReminder(ChoicePoint* rootCP);
	explicit ChoiceReminder(WorldBase const& wb, TaskList const& TL, PartialPlanAppraiser const& ppa);
	ChoiceReminder(ChoiceReminder const& cr);
	void addChoicesToCurrentNode(std::vector<ChoicePoint*> const& vec);
	bool backtrack();
	bool update();
	ChoicePoint* getRootPoint() const;
	ChoicePoint* getCurrentPoint() const;
	ChoicePoint::Tag getCurrentPointTag();
	ChoicePoint::Type getCurrentPointType();
	void setCurrentPointTag(ChoicePoint::Tag tag);
	LightWorldBase* getCurrentWorldBase();
	LightWorldBase const* getCurrentWorldBase() const;
	TaskList* getCurrentTaskList();
	TaskList const* getCurrentTaskList() const;
	std::vector<unsigned long int> const& getCurrentProjectionOrder() const;
	unsigned long int getCurrentTask() const;
	PlanTree* getCurrentPlanTree();
	PlanTree const* getCurrentPlanTree() const;
	PartialPlanAppraiser* getCurrentPartialEval();
	PartialPlanAppraiser const* getCurrentPartialEval() const;
	void selectUnexploredChildOfCurrentPoint();
	unsigned long int getNbChoicePoints() const;
	std::set<ChoicePoint*> const& getChoicePoints() const;
	std::vector<unsigned long int> getCurrentFatherChoices() const;

	std::string toString() const;

	~ChoiceReminder();
};

/**
 * return root point
 */
inline ChoicePoint* ChoiceReminder::getRootPoint() const
{
	return root_point;
}

/**
 * return current point
 */
inline ChoicePoint* ChoiceReminder::getCurrentPoint() const
{
	return current_point;
}

/**
 * return tag of the current point
 */
inline ChoicePoint::Tag ChoiceReminder::getCurrentPointTag()
{
	return current_point->getTag();
}

/**
 * return type of the current point
 */
inline ChoicePoint::Type ChoiceReminder::getCurrentPointType()
{
	return current_point->getType();
}


/**
 * calls setTag(tag) for the current node
 */
inline void ChoiceReminder::setCurrentPointTag(ChoicePoint::Tag tag)
{
	current_point->setTag(tag);
}

/**
 * return WorldBase of the current ChoicePoint
 */
inline LightWorldBase* ChoiceReminder::getCurrentWorldBase()
{
	return current_point->getWorldBase();
}

/**
 * return WorldBase of the current ChoicePoint (const version)
 */
inline LightWorldBase const* ChoiceReminder::getCurrentWorldBase() const
{
	return current_point->getWorldBase();
}

/**
 * return TaskList of the current ChoicePoint
 */
inline TaskList* ChoiceReminder::getCurrentTaskList()
{
	return current_point->getTaskList();
}

/**
 * return TaskList of the current ChoicePoint (const version)
 */
inline TaskList const* ChoiceReminder::getCurrentTaskList() const
{
	return current_point->getTaskList();
}

/**
 * return projection_order of the current ChoicePoint
 */
inline std::vector<unsigned long int> const& ChoiceReminder::getCurrentProjectionOrder() const
{
	return current_point->getProjectionOrder();
}

/**
 * return Task of the current ChoicePoint
 */
inline unsigned long int ChoiceReminder::getCurrentTask() const
{
	return current_point->getTask();
}

/**
 * return PlanTree of the current ChoicePoint
 */
inline PlanTree* ChoiceReminder::getCurrentPlanTree()
{
	return current_point->getTree();
}

/**
 * return PlanTree of the current ChoicePoint (const version)
 */
inline PlanTree const* ChoiceReminder::getCurrentPlanTree() const
{
	return current_point->getTree();
}

/**
 * return PartialPlanAppraiser of the current ChoicePoint
 */
inline PartialPlanAppraiser* ChoiceReminder::getCurrentPartialEval()
{
	return current_point->getPartialEval();
}

/**
 * return PartialPlanAppraiser of the current ChoicePoint (const version)
 */
inline PartialPlanAppraiser const* ChoiceReminder::getCurrentPartialEval() const
{
	return current_point->getPartialEval();
}

/**
 * return points
 */
inline std::set<ChoicePoint*> const& ChoiceReminder::getChoicePoints() const
{
	return points;
}

/**
 * returns the number of choice points in the choice reminder
 */
inline unsigned long int ChoiceReminder::getNbChoicePoints() const
{
	return points.size();
}

#endif
