#if !defined HATP_LIGHT_WORLDBASE_HH_
#define HATP_LIGHT_WORLDBASE_HH_

#include<string>
#include<vector>
#include<map>

#include "LightEntity.hh"
#include "WorldBase.hh"
#include "LightWorldBaseAgent.hh"

/** Represents the world as a set of LightEntity and a set of agents.
 * The agents are in fact LightWorldBaseAgent to only store their knowledge of the world.
 */
class LightWorldBase {

private:
	/** Is used when returning a reference to an entity after an error in operator[](const char* c). */
	LightEntity defaultLightEntity;
	/** Contains the set of entities in the world. */
	std::map<std::string, LightEntity*> database;

	/** Tells whether the belief management is active or not */
	bool activeBM;
	/** The name of this database, is used in LightWorldBase::findForAgent(std::string agentName, std::string entityName) */
	std::string myselfName;
	/** Is the set of agents in the world and their representation of the world (See LightWorldBaseAgent). */
	std::map<std::string, LightWorldBaseAgent*> WBAgents;

public:
	explicit LightWorldBase();
	explicit LightWorldBase(WorldBase const& wb);
	LightWorldBase(LightWorldBase const& lwb);
	std::map<std::string, LightEntity*> const& getDatabase() const;
	bool containEntity(std::string const& sName) const;
	bool containEntity(std::string const& sName, std::string const& sType) const;
	LightEntity* find(std::string const& sName);
	LightEntity const* find(std::string const& sName) const;
	LightEntity* find(std::string const& sName, std::string const& sType);
	LightEntity const* find(std::string const& sName, std::string const& sType) const;
	std::vector<LightEntity*> findByType(std::string const& sType) const;

	const bool isActiveBM() const;
	const std::string getMyselfName() const;
	const std::vector<std::string> getAllAgentsName() const;
	const std::map<std::string, LightWorldBaseAgent*>& getWBAgents() const;
	bool hasAgent(std::string const& agentName);
	LightEntity* findForAgent(std::string const& agentName, std::string const& entityName);
	void updateForAgent(std::string const& a1, std::string const& a2, std::string const& e, std::string const& attr);
	void updateForAgent(std::string const& a2, LightEntity* efa1, LightEntity* efa2, std::string const& attr);

	LightEntity* getEntity(std::string const& agentName, std::string const& entityName);
	bool updateAtomAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName, std::string const& value);//TODO make another with template parameter
	bool clearSetAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName);
	bool addToSetAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName, std::string const& value);//TODO make another with template parameter

	std::string toString() const;
	LightEntity& operator[](const char* c);
	LightEntity& operator[](std::string const& s);
	LightEntity const& operator[](const char* c) const;
	LightEntity const& operator[](std::string const& s) const;
	LightWorldBase& operator=(LightWorldBase const& lwb);
	~LightWorldBase();
	
	typedef LightEntity EntityType;
};

/** Default constructor.
 * All constructors set defaultEntity to an empty LightEntity.
 */
inline LightWorldBase::LightWorldBase() {}

/** Construct the LightWorldBase from a WorldBase.
 * All constructors set defaultEntity to an empty LightEntity.
 * @param wb World database to copy
 */
inline LightWorldBase::LightWorldBase(WorldBase const& wb) {
	std::map<Entity*, LightEntity*> oldToNew; // to keep correspondence
	LightEntity *e;
	std::map<std::string, Entity*> tmp=wb.getDatabase();
	std::map<std::string, Entity*>::const_iterator it;
	for(it = tmp.begin() ; it!=tmp.end() ; ++it) {
		e = new LightEntity(*(it->second));
		database[it->first] = e;
		oldToNew[it->second] = e;
	}
	// we need to update the LightEntity* attributes
	std::map<std::string, Entity*> atomsMap;
	std::map<std::string, Entity*>::const_iterator itAtoms;
	std::map<std::string, std::set<Entity*> > setsMap;
	std::map<std::string, std::set<Entity*> >::const_iterator itSets;
	std::set<Entity*>::const_iterator itS;
	for(it = tmp.begin() ; it!=tmp.end() ; ++it) { // for each entity in wb/tmp
		atomsMap = it->second->getEntityAttrs().getStaticAtomAttrs();
		for(itAtoms = atomsMap.begin() ; itAtoms != atomsMap.end() ; ++itAtoms) // for each attribute (static atom entity) of this entity
			database[it->first]->get<LightEntity*>(itAtoms->first) = oldToNew[itAtoms->second]; // upd entity->attr in db with new value
		atomsMap = it->second->getEntityAttrs().getDynamicAtomAttrs();
		for(itAtoms = atomsMap.begin() ; itAtoms != atomsMap.end() ; ++itAtoms)
			database[it->first]->get<LightEntity*>(itAtoms->first) = oldToNew[itAtoms->second];
		setsMap = it->second->getEntityAttrs().getStaticSetAttrs();
		for(itSets = setsMap.begin() ; itSets != setsMap.end() ; itSets++) {
			database[it->first]->get< std::set<LightEntity*> >(itSets->first).clear();
			for(itS = (itSets->second).begin() ; itS != (itSets->second).end() ; ++itS)
				database[it->first]->get< std::set<LightEntity*> >(itSets->first).insert(oldToNew[*itS]);
		}
		setsMap = it->second->getEntityAttrs().getDynamicSetAttrs();
		for(itSets = setsMap.begin() ; itSets != setsMap.end() ; itSets++) {
			database[it->first]->get< std::set<LightEntity*> >(itSets->first).clear();
			for(itS = (itSets->second).begin() ; itS != (itSets->second).end() ; ++itS)
				database[it->first]->get< std::set<LightEntity*> >(itSets->first).insert(oldToNew[*itS]);
		}
	}
	this->activeBM = wb.isActiveBM();
	this->myselfName = wb.getMyselfName();
	std::map<std::string, WorldBaseAgent*> const& atmp=wb.getWBAgents();
	for(std::map<std::string, WorldBaseAgent*>::const_iterator it = atmp.begin() ; it!=atmp.end() ; ++it) {
		this->WBAgents[it->first] = new LightWorldBaseAgent(*(it->second), this->getDatabase());
	}	
}

/** Copy constructor.
 * All constructors set defaultEntity to an empty LightEntity.
 * @param lwb World database to copy
 */
inline LightWorldBase::LightWorldBase(LightWorldBase const& lwb) {
	std::map<LightEntity*, LightEntity*> oldToNew; // to keep correspondence
	LightEntity *e;
	std::map<std::string, LightEntity*> const& tmp=lwb.getDatabase();
	for(std::map<std::string, LightEntity*>::const_iterator it = tmp.begin() ; it!=tmp.end() ; ++it) {
		e = new LightEntity(*(it->second));
		database[it->first] = e;
		oldToNew[it->second] = e;
	}
	// we need to update the Entity* attributes
	for(std::map<std::string, LightEntity*>::iterator itD = database.begin() ; itD!=database.end() ; ++itD) {	
		itD->second->updateAttrs(oldToNew);
	}
	this->activeBM = lwb.isActiveBM();
	this->myselfName = lwb.getMyselfName();
	std::map<std::string, LightWorldBaseAgent*> const& atmp=lwb.getWBAgents();
	for(std::map<std::string, LightWorldBaseAgent*>::const_iterator it = atmp.begin() ; it!=atmp.end() ; ++it) {
		this->WBAgents[it->first] = new LightWorldBaseAgent(*(it->second), oldToNew);
	} 
}


/** Gives a reference to the entity database.
 * @return LightEntity database
 */
inline std::map<std::string, LightEntity*> const& LightWorldBase::getDatabase() const {
	return database;
}

/** Tells whether the Belief Management is active or not.
 * @return True if the Belief Management is active
 */
inline const bool LightWorldBase::isActiveBM() const {
	return activeBM;
}

/** Gives the name of this database.
 * @return The name of this database
 */
inline const std::string LightWorldBase::getMyselfName() const {
	return myselfName;
}



/** Gives the list of all agents' name (except the one called myself).
 * @return List of all agents' name
 */
inline const std::vector<std::string> LightWorldBase::getAllAgentsName() const {
	std::vector<std::string> result;
	std::map<std::string, LightWorldBaseAgent*>::const_iterator it;
	for(it = WBAgents.begin() ; it != WBAgents.end() ; it++) result.push_back(it->first);
	return result;
}

/** Gives the set of LightWorldBaseAgent.
 * The map key is the agents' name.
 * @return Agent Database
 */
inline const std::map<std::string, LightWorldBaseAgent*>& LightWorldBase::getWBAgents() const {
	return WBAgents;
}

/** Find an entity in a given database.
 * @param agentName Can either be the name of this database or the name of an agent from the world
 * @param entityName The entity to find
 * @return A pointer to the entity or NULL if the agent/entity does not exist
 */
inline LightEntity* LightWorldBase::findForAgent(std::string const& agentName, std::string const& entityName) {
	if(agentName == myselfName) {
		return this->find(entityName);
	}
	if(agentName == "Myself") {
		return this->find(entityName);
	}
	if(WBAgents.find(agentName) != WBAgents.end()) {
		return WBAgents.find(agentName)->second->find(entityName);
	}
	return NULL;
}

/** Find an entity in the world database.
 * @param sName The entity to find
 * @return A pointer to the entity or NULL if the entity does not exist
 */
inline LightEntity* LightWorldBase::find(std::string const& sName) {
	std::map<std::string, LightEntity*>::iterator it = database.find(sName);
	if(it != database.end()) return it->second;
	std::cerr << "ERROR when trying to access to an unknown Entity \"" << sName;
	std::cerr << "\" in the WorldBase." << std::endl << "NULL returned instead" << std::endl;
	return NULL;
}

/** Find an entity in the world database.
 * @param sName The entity to find
 * @return A const pointer to the entity or NULL if the entity does not exist
 */
inline LightEntity const* LightWorldBase::find(std::string const& sName) const {
	std::map<std::string, LightEntity*>::const_iterator it = database.find(sName);
	if(it != database.end()) return it->second;
	std::cerr << "ERROR when trying to access to an unknown Entity \"" << sName;
	std::cerr << "\" in the WorldBase." << std::endl << "NULL returned instead" << std::endl;
	return NULL;
}

/** Assignement operator.
 * @param lwb LightWorldBase to copy
 * @return Reference to this
 */
inline LightWorldBase& LightWorldBase::operator=(LightWorldBase const& lwb) {
	// we empty database
	for(std::map<std::string, LightEntity*>::iterator itD = database.begin() ; itD != database.end() ; ++itD)
		delete itD->second;
	database.clear();
	std::map<LightEntity*, LightEntity*> oldToNew; // to keep correspondence
	LightEntity *e;
	std::map<std::string, LightEntity*> const& tmp=lwb.getDatabase();
	for(std::map<std::string, LightEntity*>::const_iterator it = tmp.begin() ; it!=tmp.end() ; ++it) {
		e = new LightEntity(*(it->second));
		database[it->first] = e;
		oldToNew[it->second] = e;
	}
	// we need to update the Entity* attributes
	for(std::map<std::string, LightEntity*>::iterator itD = database.begin() ; itD!=database.end() ; ++itD)
		itD->second->updateAttrs(oldToNew);
	// for agents
	this->activeBM = lwb.isActiveBM();
	this->myselfName = lwb.getMyselfName();
	std::map<std::string, LightWorldBaseAgent*> const& atmp=lwb.getWBAgents();
	for(std::map<std::string, LightWorldBaseAgent*>::const_iterator it = atmp.begin() ; it!=atmp.end() ; ++it) {
		this->WBAgents[it->first] = new LightWorldBaseAgent(*(it->second));
	} 
	return *this;
}

#endif
