/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_PLANNING_SESSION_HH_
#define HATP_PLANNING_SESSION_HH_

#include<cmath>
#include<algorithm>
#include<boost/date_time/posix_time/posix_time.hpp>

#include "Entity.hh"
#include "WorldBase.hh"
#include "LightWorldBase.hh"
#include "HTNBase.hh"
#include "SocialRuleBase.hh"
#include "ChoicePoint.hh"
#include "ChoiceReminder.hh"
#include "PlanStockManager.hh"
#include "PartialPlanAppraiser.hh"
#include "TaskInstance.hh"
#include "TaskListNode.hh"
#include "TaskList.hh"
#include "PlanTree.hh"
#include "Plan.hh"
#include "PlanSender.hh"
#include "MainSessionMsg.hh"
#include "WorldBaseReport.hh"

//Handle verbose level
static std::ofstream dev_null("/dev/null");
#define VERBOSE_1 (VERBOSE_LEVEL>=1?std::cout:dev_null)
#define VERBOSE_2 (VERBOSE_LEVEL>=2?std::cout:dev_null)
#define VERBOSE_3 (VERBOSE_LEVEL>=3?std::cout:dev_null)

class hatpPlanningSession
{
private:
	/**
	* the HTNBase of the planning session
	*/
	HTNBase const& htnb;

	/**
	* the SocialRuleBase of the planning session
	*/
	SocialRuleBase const& srb;

	/**
	* the original WorldBase of the planning session
	*/
	WorldBase* original_WB;

	/**
	* the ChoiceReminder of the session
	*/
	ChoiceReminder choice_tree;

	/**
	* the WorldBase of the planning session
	*/
	LightWorldBase* current_WB;

	/**
	* the current TaskList
	*/
	TaskList* current_TL;

	/**
	* the current vector defining order in which tasks have been projected
	*/
	std::vector<unsigned long int> current_projection_order;

	/**
	* the ID of the current task
	*/
	unsigned long int current_T;

	/**
	* the current PlanTree
	*/
	PlanTree* current_PT;

	/**
	* the current partial evaluation of the plan
	*/
	PartialPlanAppraiser* current_Eval;

	/**
	* true if there is a time limit to the planning session
	*/
	bool time_limit;

	/**
	* duration of the time limit in ms
	*/
	unsigned long int duration;

	/**
	* true if the process should stop at the first plan
	*/
	bool stopAtFirstPlan;

	/**
	* communicator
	*/
	PlanSender<MainSessionMsg>* communicator;

	/**
	* verbose level for degub output
	* See HATPOnboard/onboard/mainOnboardHATP.cc to have detail about meaning
	*/
	unsigned short VERBOSE_LEVEL;

	void updateCurrentData();
	float computeContribution(std::map<std::string, float> const& bestPlanData, PartialPlanAppraiser const& ppa) const;
	bool runPhase2();

public:
	explicit hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB,
		TaskList const& TL, PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock, bool argTimeLimit=false, 
		unsigned long int argDuration=0, bool stopAtFirstPlan=false, unsigned short verbose=0);
	explicit hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB,
		TaskList const& TL, PlanTree const& PT, PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock,
	bool argTimeLimit=false, unsigned long int argDuration=0, bool stopAtFirstPlan=false, unsigned short verbose=0);
	hatpPlanningSession(hatpPlanningSession const& otherSession);
	void run();
	HTNBase const& getHTNBase() const;
	SocialRuleBase const& getSocialRuleBase() const;
	WorldBase* getOriginalWorldBase() const;
	bool getTimeLimit() const;
	unsigned int getDuration() const;
	PlanSender<MainSessionMsg>* getCommunicator() const;
	ChoiceReminder& getChoiceReminder();
	ChoiceReminder const& getChoiceReminder() const;
	void operator()();
	~hatpPlanningSession();
	
	TaskInstance* getPredecessorInstance(const TaskInstance& instance) const;
	std::string toString() const;
};

/**
 * sets current_TL, current_PP and current_PT from data
 * given by current point of choice_tree (used when backtracking during run)
 */
inline void hatpPlanningSession::updateCurrentData()
{
	current_WB = choice_tree.getCurrentWorldBase();
	current_TL = choice_tree.getCurrentTaskList();
	current_projection_order = choice_tree.getCurrentProjectionOrder();
	current_T = choice_tree.getCurrentTask();
	current_PT = choice_tree.getCurrentPlanTree();
	current_Eval = choice_tree.getCurrentPartialEval();
}

/**
 * compute contribution using scoreDetails and data stored in ppa
 */
inline float hatpPlanningSession::computeContribution(std::map<std::string, float> const& bestPlanData, PartialPlanAppraiser const& ppa) const
{
	float p_pref=0., p_p=0., s_ref, s_p, N; // N is the level of preference of Pref instead of P
	std::map<std::string, float>::const_iterator it, itW, itV;
	std::map<std::string, float> const& badDecompoPen = ppa.getBadDecompoPenalties();
	std::map<std::string, float> const& undSeqPen = ppa.getUndSeqPenalties();
	std::map<std::string, float> const& weights = Plan::scoreWeights;
	for(it = bestPlanData.begin() ; it != bestPlanData.end() ; ++it) {
		s_ref = it->second;
		if(it->first == COSTS_PART) {
			s_p = ppa.getCurrentProjectionCost();
			if(s_p > s_ref) {
				N = round(10.*(s_p - s_ref)/s_ref);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
			}
			else if(s_p == s_ref) N = 1.;
			else {
				N = round(10.*(s_ref - s_p)/s_ref);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
				N = 1./N;
			}
		}
		else {
			itV = badDecompoPen.find(it->first);
			if(itV != badDecompoPen.end()) s_p = itV->second;
			else {
				itV = undSeqPen.find(it->first);
				if(itV != undSeqPen.end()) s_p = itV->second;
				else s_p = 0.;// we consider that other criteria are equal to 0 in the current plan
			}
			itW = weights.find(it->first);
			if(s_p > s_ref) {
				N = round(s_p-s_ref);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
			}
			else if (s_p == s_ref) N = 1.;
			else {
				N = round(s_ref-s_p);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
				N = 1./N;
			}
		}
		itW = weights.find(it->first);
		p_pref += N/(N+1.) * (itW->second);
		p_p += 1./(N+1.) * (itW->second);
	}
	return p_pref-p_p; // if returned number > 0, we will apply branch and bound
}

/**
 * return htnb
 */
inline HTNBase const& hatpPlanningSession::getHTNBase() const
{
	return htnb;
}

/**
 * return srb
 */
inline SocialRuleBase const& hatpPlanningSession::getSocialRuleBase() const
{
	return srb;
}

/**
 * return original_WB
 */
WorldBase* hatpPlanningSession::getOriginalWorldBase() const
{
	return original_WB;
}

/**
 * return time_limit
 */
inline bool hatpPlanningSession::getTimeLimit() const
{
	return time_limit;
}

/**
 * return duration
 */
inline unsigned int hatpPlanningSession::getDuration() const
{
	return duration;
}

/**
 * return communicator
 */
inline PlanSender<MainSessionMsg>* hatpPlanningSession::getCommunicator() const
{
	return communicator;
}

/**
 * return choice_tree
 */
inline ChoiceReminder& hatpPlanningSession::getChoiceReminder()
{
	return choice_tree;
}

/**
 * return choice_tree (const version)
 */
inline ChoiceReminder const& hatpPlanningSession::getChoiceReminder() const
{
	return choice_tree;
}

#endif
